from pyqalx.factories.sectors.providers.aws.provider import AWSProvider
from pyqalx.factories.sectors.providers.local.provider import LocalProvider

# A dict of valid external providers.
_providers = {
    AWSProvider._type: AWSProvider,
    LocalProvider._type: LocalProvider,
}



import json
from collections import defaultdict

import yaml
from troposphere import Tag, Tags, Template, Base64, Join, Ref, ssm, cloudwatch
from troposphere.ec2 import Instance
from troposphere.iam import Role, Policy, InstanceProfile
from troposphere.secretsmanager import Secret
from troposphere.cloudformation import WaitCondition, WaitConditionHandle

from pyqalx.__version__ import __version__
from pyqalx.config import UserConfig, BotConfig


class AWSStack:
    def __init__(self, provider):
        self.provider = provider
        self.factory_entity = self.provider.factory.entity
        self.stack = self.provider.stack
        self.template = Template(Description=self.provider._description)
        self.images = None
        self._secure_parameters = ["TOKEN"]

    @staticmethod
    def _add_tag(key, value):
        """
        Helper for adding a Tag to a resource
        :param key: The key for the tag
        :type key: str
        :param value: The value for the tag
        :type value: str
        :return:dict
        """
        return Tag(key, value)

    def _resource_tags(self, name):
        """
        Helper for common resource tags to use on resources created by
        the stack, as well as the stack itself
        """
        tags = [
            self._add_tag("Name", name),
            # This should be stored as a string on AWS not uuid.UUID
            self._add_tag("qalx-factory-guid", str(self.factory_entity.guid)),
            self._add_tag("qalx-factory-stage", self.provider.stage),
            self._add_tag("qalx-created-version", __version__),
            self._add_tag("qalx-factory-name", self.factory_entity.name),
            self._add_tag(
                "qalx-creating-aws-profile", self.provider.profile_name
            ),
        ]
        return tags

    def _tags(self, sector_name, suffix):
        name = self.provider._resource_name(sector_name, suffix)
        tags = self._resource_tags(name=name)
        return Tags(*tags)

    def _tags_as_dict(self, sector_name, suffix):
        """
        For AWS reasons some resources require the Tags to be a dict of
        {key:value} rather than a list of [{"Key": key, "Value": value}]
        """
        _tags = self._tags(sector_name, suffix).to_dict()
        tags = {}
        for tag in _tags:
            tags[tag["Key"]] = tag["Value"]
        return tags

    def _logical_id(self, sector_name, suffix):
        return (
            self.provider._resource_name(sector_name, suffix)
            .title()
            .replace("-", "")
        )

    def _get_instance_name(self):
        """
        The instance name only exists in the tags of the instance.
        It will always be there as the instance gets created before the
        secrets
        """
        ec2_instance = self.refs["instance"]
        for tag in ec2_instance.Tags.to_dict():
            if tag["Key"] == "Name":
                instance_name = tag["Value"]
                break
        return instance_name

    def _parameter_variable_name(self, name, key_prefix):
        """
        Parameter variables don't follow the same naming convention
        as normal resources.  Instead they should be in the format of
        "<instance_name>_<config prefix><variable name>".
        This enables easier lookup of the resource in order to set them
        as environment variables on the EC2 instance.  It also allows separate
        profiles to be used for different EC2 instances in the stack
        :param name:The name of the paramter variable we want to save
        :return: The full name of the parameter variable how it should exist
        on the EC2 instance
        """
        instance_name = self._get_instance_name()
        return f"{instance_name}_{key_prefix}{name.upper()}"

    def _parameter_description(self, name):
        return f"qalx {name.upper()} for {self._get_instance_name()}"

    def linux_user_data(self, sector_name):
        """
        The user data that enables a linux instance to be
        bootstrapped.
        :param sector_name:The name of the sector
        :return: troposphere template detailing the userdata
        """
        sector = self.provider.stack[sector_name]
        instance_name = self._get_instance_name()
        bootstrap_script = (
            f"sudo -u ec2-user bash -c "
            f"'. /home/ec2-user/pyqalx/pyqalx_factory_bootstrap.sh "
            f"--factory_guid {self.factory_entity.guid} "
            f"--sector_name {sector_name} "
            f"--profile_name {sector['profile_name']} "
            f"--region_name {sector['region_name']} "
            f"--instance_name {instance_name}'"
        )
        # A log file the bootstrap script writes to disk that is used to check
        # for success.
        bootstrap_log = "/tmp/bootstrap.log"
        return Base64(
            Join(
                "",
                [
                    # https://aws.amazon.com/premiumsupport/knowledge
                    #           -center/execute-user-data-ec2/
                    'Content-Type: multipart/mixed; boundary="//"\n',
                    "MIME-Version: 1.0\n\n",
                    "--//\n",
                    'Content-Type: text/cloud-config; charset="us-ascii"\n',
                    "MIME-Version: 1.0\n",
                    "Content-Transfer-Encoding: 7bit\n"
                    'Content-Disposition: attachment; filename="cloud-config.txt"\n\n'
                    "#cloud-config\n"
                    "cloud_final_modules:\n"
                    "- [scripts-user, always]\n\n"
                    "--//\n",
                    'Content-Type: text/x-shellscript; charset="us-ascii"\n',
                    "MIME-Version: 1.0\n"
                    "Content-Transfer-Encoding: 7bit\n"
                    'Content-Disposition: attachment; filename="userdata.txt"\n\n'
                    "#!/bin/bash\n",
                    # can't use cfn-init because cfn-init waits until spawned
                    # processes exit - which won't work for us because the bots
                    # will run indefinitely.
                    bootstrap_script,
                    "\n",
                    # Read the first char from the bootstrap script.  This
                    # will be 0 if success, or 1 if failure
                    f'/opt/aws/bin/cfn-signal -e "$(head -c 1 {bootstrap_log})" -r '
                    f'"$(<{bootstrap_log})" \'',
                    Ref(self.refs["wait_handle"]),
                    "'\n",
                    "-- //\n",
                ],
            )
        )

    def windows_user_data(self, sector_name):
        sector = self.provider.stack[sector_name]
        instance_name = self._get_instance_name()
        cmd = (
            "C:\\Users\\Administrator\\.venvs\\pyqalx\\Scripts\\pyqalx_"
            "factory_bootstrap.bat"
        )
        bootstrap_script = (
            f"{cmd} "
            f"--factory_guid {self.factory_entity.guid} "
            f"--sector_name {sector_name} "
            f"--profile_name {sector['profile_name']} "
            f"--region_name {sector['region_name']} "
            f"--instance_name {instance_name}"
        )

        return Base64(
            Join(
                "",
                [
                    "<script>",
                    "\n",
                    bootstrap_script,
                    "\n",
                    "</script>",
                    "\n",
                    "<powershell>",
                    "\n",
                    # A log file the bootstrap script writes to disk that is
                    # used to check for success.
                    "$logfile=Get-Content -Path C:\\Windows\\Temp\\bootstrap.log",
                    "\n",
                    "$errorcode=$logfile.SubString(0,1)",
                    "\n",
                    'cfn-signal -e "$errorcode" -r "$logfile" \'',
                    Ref(self.refs["wait_handle"]),
                    "'\n",
                    "</powershell>",
                    "\n",
                    "<persist>true</persist>",
                    "\n",
                ],
            )
        )

    def _get_platform(self, sector_name):
        """
        Works out the platform for the requested image required by the sector
        based on the ImageId.  If it is a linux image then the `Platform` key
        will not exist in the `self.images` response from AWS.
        :param sector_name:The name of the sector we are building the image for
        :return:The platform - either 'windows' or 'linux'.
        """
        sector_instance_id = self.stack[sector_name]["parameters"]["ImageId"]
        platform = None
        for image in self.images:
            if sector_instance_id == image["ImageId"]:
                # If this is a linux image the Platform key won't exist
                platform = image.get("Platform", "linux").lower()
        return platform

    def _get_user_data(self, sector_name):
        """
        Gets the user data based on the platform
        """
        platform = self._get_platform(sector_name)
        user_data = getattr(self, f"{platform}_user_data")(sector_name)
        return user_data

    def _ec2_instance_logical_id(self, sector_name):
        """
        Helper for determining the EC2 instance logical ID for a given
        sector
        :param sector_name:The name of the sector from the plan
        :return: The logical ID used in CF templates
        """
        return self._logical_id(sector_name, "instance")

    def _ec2_instance(self, sector_name):
        """
        Adds a single ec2 instance to the template.  There will be one EC2
        instance for each sector on the stack.  Creates a wait handle so that
        the instance can single to cloudformation when the bots have been
        bootstrapped (either success or failure)
        :param sector_name: The name of the sector.
        """
        suffix = "instance"
        instance_profile = self.refs["instance_profile"]
        logical_id = self._ec2_instance_logical_id(sector_name)
        wait_handle = WaitConditionHandle(
            self._logical_id(sector_name, "wait_condition_handle")
        )

        self.refs["wait_handle"] = wait_handle
        self.template.add_resource(wait_handle)
        instance = Instance(
            logical_id,
            IamInstanceProfile=instance_profile.InstanceProfileName,
            Tags=self._tags(sector_name, suffix),
            DependsOn=[self.refs["role"].title, instance_profile.title],
        )
        self.refs["instance"] = instance
        # user data needs the instance name, so set user data after the instance
        # resource is crated
        instance.UserData = self._get_user_data(sector_name)

        wait_condition = WaitCondition(
            self._logical_id(sector_name, "wait_condition"),
            DependsOn=[instance],
            Count=1,
            Handle=Ref(wait_handle),
            # Really long timeout to give time for windows instances to fully
            # boot up
            Timeout=1200,
        )
        self.template.add_resource(wait_condition)
        self.template.add_resource(instance)

    def _role_policies(self):
        """
        Updates the role with policies that limit the reading of
        secrets/parameters to just sector the role is attached to
        """
        instance_name = self._get_instance_name()
        self.refs["role"].Policies = [
            Policy(
                PolicyName="ReadSecretsManagerPolicy",
                PolicyDocument={
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Effect": "Allow",
                            "Action": "secretsmanager:GetSecretValue",
                            "Resource": {
                                "Fn::Join": [
                                    "",
                                    [
                                        f"arn:aws:"
                                        f"secretsmanager:"
                                        f"{self.provider.region_name}:",
                                        {"Ref": "AWS::AccountId"},
                                        f":secret:{instance_name}*",
                                    ],
                                ]
                            },
                        },
                        {
                            "Effect": "Allow",
                            "Action": "secretsmanager:ListSecrets",
                            "Resource": "*",
                        },
                    ],
                },
            ),
            Policy(
                PolicyName="ReadSSMPolicy",
                PolicyDocument={
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Effect": "Allow",
                            "Action": "ssm:GetParameter",
                            "Resource": {
                                "Fn::Join": [
                                    "",
                                    [
                                        f"arn:aws:"
                                        f"ssm:"
                                        f"{self.provider.region_name}:",
                                        {"Ref": "AWS::AccountId"},
                                        f":parameter/{instance_name}*",
                                    ],
                                ]
                            },
                        },
                        {
                            "Effect": "Allow",
                            "Action": "ssm:DescribeParameters",
                            "Resource": "*",
                        },
                    ],
                },
            ),
        ]

    def _role(self, sector_name):
        """
        The Role that this specific EC2 instance will have applied to it.
        Required so that the EC2 instance get obtain the configuration
        parameters from SSM and Secrets Manager.   There is one role
        per sector to ensure that sectors (EC2 instances) cannot get
        resources assigned to another sector
        """
        suffix = "role"
        role_name = self.provider._resource_name(sector_name, suffix)
        instance = Role(
            self._logical_id(sector_name, suffix),
            # role name is limited to 64 characters
            RoleName=role_name[0:64],
            Description="IAM role for qalx factory",
            AssumeRolePolicyDocument={
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {"Service": ["ec2.amazonaws.com"]},
                        "Action": ["sts:AssumeRole"],
                    }
                ],
            },
            Tags=self._tags(sector_name, suffix),
        )
        self.template.add_resource(instance)
        self.refs["role"] = instance

    def _instance_profile(self, sector_name):
        """
        The instance profile for the role.  One instance profile is created per
        stack - namespaced to the AWS profile
        """
        suffix = "instanceprofile"
        instance = InstanceProfile(
            self._logical_id(sector_name, suffix),
            InstanceProfileName=self.provider._resource_name(
                sector_name, suffix
            ),
            Roles=[self.refs["role"].RoleName],
            DependsOn=[self.refs["role"].title],
        )
        self.template.add_resource(instance)
        self.refs["instance_profile"] = instance

    def _config_parameters(self, sector_name):
        """
        Creates entries in Secret Manager and SSM Parameter store for the
        provided configs specified on the sector
        :param sector_name: The name of the sector
        """
        user_config = UserConfig._load_inifile(
            self.provider.stack[sector_name]["user_profile"]
        )
        bot_config = BotConfig._load_inifile(
            self.provider.stack[sector_name]["bot_profile"]
        )
        configs = {UserConfig: user_config, BotConfig: bot_config}
        for Config, config in configs.items():
            for name, value in config.items():
                # Values must be a string.  The session Config
                # will handle converting to boolean as necessary.
                value = str(value)
                suffix = f"{Config.key_prefix.lower()}{name.lower()}"
                logical_id = self._logical_id(sector_name, suffix)
                description = self._parameter_description(name)
                parameter_name = self._parameter_variable_name(
                    name, Config.key_prefix
                )

                if value:
                    if name in self._secure_parameters:
                        # Ensure we only create secrets if there is a value
                        suffix = f"secretsmanager{suffix}"
                        instance = Secret(
                            logical_id,
                            Name=parameter_name,
                            SecretString=value,
                            Description=description,
                            Tags=self._tags(sector_name, suffix),
                        )
                        self.template.add_resource(instance)
                    else:
                        # TOKEN should not be stored in SSM as it isn't secure.
                        suffix = f"ssm{suffix}"

                        instance = ssm.Parameter(
                            logical_id,
                            Type="String",
                            Value=value,
                            Name=parameter_name,
                            Description=description,
                            # ...Tags for ssm.Parameter are a dict rather
                            # than a list
                            Tags=self._tags_as_dict(sector_name, suffix),
                        )
                        self.template.add_resource(instance)

    def _build_alarm(self, alarm_name, region_name):
        """
        Helper for building a specific alarm based on the alarm name
        """
        alarm = self.provider.factory.plan["factory"]["alarms"][alarm_name]
        template = {}

        for key, value in alarm.items():
            if key == "AlarmActions":
                # Alarm actions get built up from an enum
                action_value = []
                for action in value:
                    action_value.append(
                        f"arn:aws:automate:{region_name}:ec2:{action}"
                    )
                value = action_value
            template[key] = value
        return template

    def _alarms(self):
        """
        Builds all the Cloudwatch alarms for all the instances.  The same alarm
        is used for multiple instances if possible.
        """
        alarms = {}
        dimensions = defaultdict(list)
        depends_on = defaultdict(list)
        for sector_name, parameters in self.stack.items():
            alarm_name = parameters.get("alarm", None)
            ec2_instance_logical_id = self._ec2_instance_logical_id(sector_name)
            sector = self.provider.stack[sector_name]
            region = sector["region_name"]

            if alarm_name is not None:
                # There is an alarm on this sector, build the base alarm
                # template
                logical_id = self._logical_id(
                    f"{sector_name}-{alarm_name}", "alarm"
                )
                alarms[logical_id] = self._build_alarm(alarm_name, region)
                depends_on[logical_id].append(ec2_instance_logical_id)
                dimensions[logical_id].append(
                    {
                        "Name": "InstanceId",
                        "Value": Ref(ec2_instance_logical_id),
                    }
                )

        for logical_id, dimension in dimensions.items():
            metric_dimensions = [
                cloudwatch.MetricDimension().from_dict(d=d, title=logical_id)
                for d in dimension
            ]

            alarms[logical_id]["Dimensions"] = metric_dimensions
            alarms[logical_id]["DependsOn"] = depends_on[logical_id]

            instance = cloudwatch.Alarm(logical_id, **alarms[logical_id])
            self.template.add_resource(instance)

    def _image_ids(self):
        """
        Returns the imageIDs for the given sector.  There may be multiple
        EC2 instances for a given sector
        :return: List of image ids
        """
        image_ids = []
        for sector in self.stack.values():
            image_ids.append(sector["parameters"]["ImageId"])
        return image_ids

    def generate(self):
        self.images = self.provider._describe_images(self._image_ids())

        if self.images is None:
            # There was a problem describing images.  We can't build this
            # stack so return. The error message is shown to the user
            # in the _describe_images call
            return None

        found = [i["ImageId"] for i in self.images]
        missing = set(self._image_ids()) - set(found)
        if missing:
            # The imageIds in the plan don't exist on AWS
            self.provider._msg_user(
                f"Could not find the following ImageIds on "
                f'AWS: {", ".join(missing)}.'
            )
            return None

        for sector_name in self.stack.keys():
            self.refs = {}
            # Create one role & instance profile per sector.
            # This ensures that an instance cannot talk to resources
            # specified for another instance
            self._role(sector_name)
            self._instance_profile(sector_name)
            # Create one ec2 instance and associated parameters for each sector
            self._ec2_instance(sector_name)
            self._config_parameters(sector_name)
            self._role_policies()
        self._alarms()
        # The user supplies the raw parameters for the EC2 instances
        # so we need to insert them into the template directly
        template = self._insert_parameters()
        return yaml.safe_dump(template)

    def _insert_parameters(self):
        """
        Inserts the parameters from the sector directly into the EC2
        instance properties
        :return: The template as a dict
        """
        template = json.loads(self.template.to_json())
        for sector_name, parameters in self.stack.items():
            logical_id = self._logical_id(sector_name, "instance")

            template["Resources"][logical_id]["Properties"].update(
                parameters["parameters"]
            )
        return template



from pyqalx.core.adapters.adapter import (
    QalxNamedEntityAdapter,
    QalxUpdateStatusAdapter,
    QalxUpdateStateAdapter,
)
from pyqalx.core.entities import Bot
from pyqalx.core.enums import ValidStates


class QalxBot(
    QalxNamedEntityAdapter, QalxUpdateStatusAdapter, QalxUpdateStateAdapter
):
    """
    Provides an interface for accessing the API for
    :class:`~pyqalx.core.entities.bot.Bot` entities
    """

    _entity_class = Bot
    _user_only_methods = ["add"]
    _bot_only_methods = ["replace_workers"]

    def add(self, name, config, meta=None, **kwargs):
        """
        Creates a `Bot` instance.

        :param name: The name that this bot will be given
        :type name: str
        :param config: The bots config
        :type config: dict
        :param meta: A dictionary of metadata to store
        :type meta: dict
        :return: The newly created :class:`~pyqalx.core.entities.bot.Bot` instance
        """
        return super(QalxBot, self).add(
            host=self.session._host_info,
            name=name,
            meta=meta,
            config=config,
            **kwargs,
        )

    def get_by_name(self, name, **kwargs):
        # extended to create better autodocs
        return super(QalxBot, self).get_by_name(name=name, **kwargs)

    def get_or_create(self, name, meta=None, **kwargs):
        # extended to create better autodocs
        return super(QalxBot, self).get_or_create(
            name=name, meta=meta, **kwargs
        )

    def get(self, guid, *args, **kwargs):
        # extended to create better autodocs
        return super(QalxBot, self).get(guid, *args, **kwargs)

    def save(self, entity, *args, **kwargs):
        # extended to create better autodocs
        return super(QalxBot, self).save(entity, *args, **kwargs)

    def find(
        self,
        query=None,
        sort=None,
        skip=0,
        limit=25,
        many=True,
        include_session_tags=True,
        **kwargs,
    ):
        # extended to create better autodocs
        return super(QalxBot, self).find(
            query=query,
            sort=sort,
            skip=skip,
            limit=limit,
            many=many,
            include_session_tags=include_session_tags,
            **kwargs,
        )

    def find_one(self, query, **kwargs):
        # extended to create better autodocs
        return super(QalxBot, self).find_one(query=query, **kwargs)

    def reload(self, entity, *args, **kwargs):
        # extended to create better autodocs
        return super(QalxBot, self).reload(entity=entity, **kwargs)

    def aggregate(self, aggregate, **kwargs):
        # extended to create better autodocs
        return super(QalxBot, self).aggregate(aggregate=aggregate, **kwargs)

    def replace_workers(self, bot_entity, workers):
        """
        Completely replaces any Workers on the given bot.  Will return the
        replaced workers in an unpacked state

        :param bot_entity: The Bot instance that is being changed
        :type bot_entity: :class:`~pyqalx.core.entities.bot.Bot`
        :param workers: The number of workers that this bot should have
        :type workers: int
        :return: A :class:`~pyqalx.core.entities.bot.Bot` instance with
                 the updated workers
        """
        guid = bot_entity["guid"]
        detail_endpoint = self.detail_endpoint(guid=guid)
        endpoint = f"{detail_endpoint}/replace-workers"
        self.session.log.debug(
            f"replace workers on `{self.entity_class.entity_type}` with"
            f" guid `{guid}` with `/{endpoint}`"
        )

        resp = self._process_api_request("patch", endpoint, workers=workers)
        return self._process_api_response(resp)

    def _signal_workers(self, entity, signal_method, *args, **kwargs):
        """
        Helper method for calling signals on workers.  A bot doesn't have
        the concept of `signals`.  Instead, we call a signal method we just
        apply this to all the workers on the bot
        """

        for worker in entity.get("workers", []):
            worker_method = getattr(self.session.worker, signal_method)
            worker_method(worker, bot_entity=entity, *args, **kwargs)

    def stop(self, entity, **kwargs):
        """
        Stops all the workers on the bot

        :param entity: Bot instance
        :type entity: :class:`~pyqalx.core.entities.bot.Bot`
        """
        self.update_state(entity, ValidStates.STOPPED)
        self._signal_workers(entity, signal_method="stop", **kwargs)

    def resume(self, entity, **kwargs):
        """
        Resumes all the workers on the bot

        :param entity: Bot instance
        :type entity: :class:`~pyqalx.core.entities.bot.Bot`
        """
        # Update to `idle` - if there are extant jobs the state will be
        # updated to active by the worker at the next process function
        self.update_state(entity, ValidStates.IDLE)
        self._signal_workers(entity, signal_method="resume", **kwargs)

    def terminate(self, entity, *args, **kwargs):
        """
        Terminates all the workers on the bot

        :param entity: Bot instance
        :type entity: :class:`~pyqalx.core.entities.bot.Bot`
        """
        self.update_state(entity, ValidStates.TERMINATED)
        self._signal_workers(entity, signal_method="terminate", *args, **kwargs)

    def update_status(self, entity, status, **kwargs):
        # extended to create better autodocs
        return super(QalxBot, self).update_status(
            entity=entity, status=status, **kwargs
        )



import configparser
import os
from inspect import ismodule
from os import environ
from .types import (
    BOOL,
    LOGGING_OPTIONS,
    DIR_PATH_NON_EXISTING,
    FILE_PATH_NON_EXISTING,
    FILE_PATH_EXISTING,
    POSITIVE_INT,
)

from click.types import STRING
from click import BadParameter

from . import defaults
from pyqalx.config.defaults import bots
from pyqalx.core.errors import (
    QalxConfigProfileNotFound,
    QalxConfigFileNotFound,
    QalxConfigError,
)


class Config(dict):
    key_prefix = ""
    filename = ""
    default = "default"
    default_type_map = {
        "TOKEN": STRING,
        "KEYFILE": FILE_PATH_EXISTING,
        "LOGGING_LEVEL": LOGGING_OPTIONS,
        "LOG_FILE_PATH": FILE_PATH_NON_EXISTING,
        "MSG_BLACKOUTSECONDS": POSITIVE_INT,
        "MAX_RETRY_500": POSITIVE_INT,
        "FACTORY_PACK_DIR": DIR_PATH_NON_EXISTING,
        "DOWNLOAD_DIR": DIR_PATH_NON_EXISTING,
        "REGION": STRING,
    }

    def __init__(self):
        dict.__init__(self, self.defaults)
        for key, value in [
            (k, v) for k, v in environ.items() if k.startswith(self.key_prefix)
        ]:
            self[key.replace(self.key_prefix, "")] = value
        self._validate(self)

    @classmethod
    def config_path(cls):
        return os.path.join(os.path.expanduser("~"), cls.filename)

    @classmethod
    def _load_inifile(cls, profile_name):
        """
        Loads the inifile and returns the values of the specific
        profile_name as a dict
        :param profile_name: The name of the profile to load
        :return: Dict of config options
        :raises: QalxConfigProfileNotFound
        """
        if os.path.exists(cls.config_path()):
            config = configparser.ConfigParser()
            config.optionxform = str
            with open(cls.config_path()) as cfg:
                config.read_string(
                    cfg.read()
                )  # this makes mocks easier over config.read(filepath)
            if profile_name in config.keys():
                config_dict = config[profile_name]
            else:
                profiles = "\n".join(config.keys())
                msg = (
                    "Couldn't find profile named {} in {} file."
                    " Did you mean one of:\n{}".format(
                        profile_name, cls.filename, profiles
                    )
                )
                raise QalxConfigProfileNotFound(msg)
            return dict(config_dict)
        else:
            raise QalxConfigFileNotFound(
                "Couldn't find {}".format(cls.config_path())
            )

    def from_inifile(self, profile_name=None):
        if profile_name is None:
            profile_name = self.default
        config_dict = self._load_inifile(profile_name=profile_name)
        self.update(config_dict)
        self._validate(self)

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, dict.__repr__(self))

    @property
    def defaults(self):
        """
        Returns a dict of default values for config options from
        `~pyqalx.config.defaults.__init__`

        :return : dict of options
        """
        return {
            k: v
            for k, v in vars(defaults).items()
            if (not k.startswith("__")) and (not ismodule(v))
        }

    @classmethod
    def configure(cls, profile_name, config_items=None):
        """
        When given a profile name and dict of config items will write the
        config file to disk.  If a `profile_name` is given that already exists
        on the profile then the `profile_name` on the profile will be
        completely replaced with the values from config_items

        :param profile_name: The name of the profile to write
        :type profile_name: str
        :param config_items: A dict of items that should be written to the
                             config
        :type config_items: dict
        """
        if config_items is None:
            # We might just be writing a blank config (i.e. for `.bots`)
            config_items = {}
        validated_items = cls._validate(config_items)
        with open(cls.config_path(), "a") as cfg:
            config_string = f"[{profile_name}]\n"
            for config_key, config_value in validated_items.items():
                config_string += f"{config_key}={config_value}\n"
            cfg.write(config_string)

    @staticmethod
    def _get_logging_details():
        raise QalxConfigError("Must be implemented by implementing class")

    def _create_log_directories(self):
        """
        Creates the directory structure for the log files.  Iterates through
        a list of data from `_get_logging_details()` to handle different types
        of logging directories for each config
        """
        for logging_details in self._get_logging_details():
            # expects_file, key, dir_path = self._get_logging_details()
            CONFIG_KEY = logging_details["CONFIG_KEY"]
            logger_path = self[CONFIG_KEY]
            ISFILE = logging_details["ISFILE"]
            method = "isdir" if ISFILE else "isfile"
            if getattr(os.path, method)(logger_path):
                # If its a file, check it isn't already on disk as a directory,
                # and vice-versa
                msg = (
                    f"`{CONFIG_KEY}`: {logger_path} exists, but it is"
                    f" not a {'file' if ISFILE else 'directory'}."
                )
                raise QalxConfigError(msg)
            if ISFILE:
                logger_path = os.path.dirname(logger_path)
            try:
                os.makedirs(logger_path, exist_ok=True)
            except (OSError, PermissionError):
                msg = (
                    f"`{CONFIG_KEY}`: {logger_path}"
                    f" does not exist and it could not be created."
                )
                raise QalxConfigError(msg)

    @classmethod
    def _validate(cls, config_dict):
        """
        Validate and parse the values for the default config keys according to
        their respective type. For the provided config_dict with raw
        values, these are parsed and replaced
        """
        for config_key, config_value in config_dict.items():
            parse_type = cls.default_type_map.get(config_key, None)
            if parse_type is not None:
                try:
                    parsed_value = parse_type(config_value)
                except BadParameter as exc:
                    message = (
                        "There was an error parsing the config value for "
                        + f"{config_key}:\n\n{exc}"
                    )
                    raise QalxConfigError(message)
                else:
                    config_dict[config_key] = parsed_value
        return config_dict


class BotConfig(Config):
    """
    Works exactly like a dict but provides ways to fill it from ini files
    and environment variables.  There are two common patterns to populate the
    config.

    Either you can add them to the `.bots` file:

    >>> bot.config.from_botsfile(profile_name=BotConfig.default)

    Or alternatively you can define the configuration from environment variables
    starting with `QALX_BOT_`. These will be populated automatically
    but values defined in `.bots` will overwrite these if
    `bot.config.from_botsfile()` is called.

    To set environment variables before launching the application you have to set this
    environment variable to the name and value you want to use.
    On Linux and OS X use the export statement::

    >>> export QALX_BOT_LICENCE_FILE='/path/to/licence/file'

    On windows use `set` instead.
    """

    key_prefix = "QALX_BOT_"
    filename = ".bots"
    default_type_map = dict(Config.default_type_map)
    default_type_map.update(
        {
            "WORKER_LOG_FILE_DIR": DIR_PATH_NON_EXISTING,
            "SAVE_INTERMITTENT": BOOL,
            "SKIP_FINAL_SAVE": BOOL,
            "KILL_AFTER": POSITIVE_INT,
            "Q_WAITTIMESECONDS": POSITIVE_INT,
            "Q_MSGBATCHSIZE": POSITIVE_INT,
            "Q_WAITTIMEMAXSPREADING": POSITIVE_INT,
            "MSG_WAITTIMESECONDS": POSITIVE_INT,
        }
    )

    @property
    def defaults(self):
        """
        Loads the default bot config from `~pyqalx.config.defaults.bots`.  Any
        config items that exists in `~pyqalx.config.defaults.__init__` will be
        overwritten by config values that exist in `~pyqalx.config.defaults.bots`

        :return: dict of default config options
        """
        config = super(BotConfig, self).defaults
        bot_config = {
            k: v
            for k, v in vars(bots).items()
            if not k.startswith("__") and (not ismodule(v))
        }
        config.update(bot_config)
        # `REGION` will come from the super() class defaults.  However, it isn't
        # a valid option when creating a `BotConfig`, so remove it from
        # the defaults
        config.pop("REGION", None)
        return config

    @staticmethod
    def _get_logging_details():
        return [{"CONFIG_KEY": "WORKER_LOG_FILE_DIR", "ISFILE": False}]

    @classmethod
    def _validate(cls, config_dict):
        config_dict = super(BotConfig, cls)._validate(config_dict)
        if config_dict.get("REGION"):
            # To avoid confusing issues with a Bot config looking at a
            # different region to a user config, the ability to set the region
            # on a bot config inifile is prohibited.  The region can only be
            # set manually to help ensure that the region on the bot qalx session
            # matches the region on the user qalx session when initialising Bots
            raise QalxConfigError(
                "`REGION` is not a valid config option "
                f"for `{cls.__name__}`.  Did you mean to add"
                " `REGION` to a user config?"
            )
        return config_dict


class UserConfig(Config):
    """Works exactly like a dict but provides ways to fill it from ini files
    and environment variables.  There are two common patterns to populate the
    config.
    Either you can add them to the `.qalx_bot` file:

    >>> qalx_bot.config.from_qalxfile(profile_name=UserConfig.default)

    Or alternatively you can define the configuration from environment variables
    starting with `QALX_USER_`. These will be populated automatically
    but values defined in `.qalx` will overwrite these if
    `qalx_bot.config.from_qalxfile()` is called.

    To set environment variables before launching the application you have to set this
    environment variable to the name and value you want to use.
    On Linux and OS X use the export statement::

    >>>export QALX_USER_EMPLOYEE_NUMBER=1280937

    On windows use `set` instead.
    """

    key_prefix = "QALX_USER_"
    filename = ".qalx"
    # The default region to use when prompting the user as part of `qalx configure`
    DEFAULT_REGION = "london"
    # The region mapping to use when validating the region chosen by the user
    # Nested under `aws` to make it easier to add different providers in the
    # future
    REGION_MAPPING = {"aws": {DEFAULT_REGION: "eu-west-2"}}

    @staticmethod
    def _get_logging_details():
        return [{"CONFIG_KEY": "LOG_FILE_PATH", "ISFILE": True}]



"""
Factory module. Contains the main factory class
"""
import logging

from botocore.exceptions import ClientError

from .mixins import BaseFactory
from ..core.errors import (
    QalxFactoryPackError,
    QalxAPIResponseError,
    QalxFactoryDemolishError,
    QalxFactoryBuildError,
)


class Factory(BaseFactory):
    """
    Main factory class. Inherits from BaseFactory and contains methods for
    specific functionality related to the distinct stages of a factory
    """

    DEPLOYING_STATUS = "deploying"
    DEPLOYED_STATUS = "deployed"
    DELETING_STATUS = "deleting"

    def validate(self, plan):
        """
        Calls the validate method on the adapter

        :param plan: Path to a .yaml file containing the definition of the factory plan
        :type plan: str
        :raises QalxFactoryValidationError: If there is a problem validating the plan
        """
        self._msg_user(msg="Starting Factory Validation")
        self.session.factory.validate(plan)
        self.plan = self.read_plan(plan)
        self._msg_user(msg="Factory Validation Successful")

    def pack(self, plan, stage=None, delete=True):
        """
        Packs a factory. Calls validate method on the plan provided

        :param plan:    Path to a .yaml file containing the definition of the
                        factory plan
        :type plan: str
        :param stage:   Stage to be packed. This argument is optional. If None
                        is provided, then all stages will be packed
        :type stage: str
        :param delete:  Should the packed code be deleted when this command
                        completes?
        :type delete: bool

        :raises QalxFactoryPackError:   If a stage is provided and it cannot be
                                        found on the plan, or if there are any
                                        issues creating the build directory
        """
        self._msg_user(msg="Starting Factory Pack")

        self.validate(plan)

        if stage and stage not in self._plan_stages:
            raise QalxFactoryPackError(
                f"Stage {stage} was not found in the factory plan"
            )
        self._create_build_path()

        bots = self._bots_for_pack(stage=stage)
        self._download_bots(bots)
        if delete:
            self._delete_build_path()
        self._msg_user(msg="Factory Pack Successful")

    def build(self, plan, stage):
        """
        Builds a factory. Calls pack method with the provided plan and stage.

        :param plan:    Path to a .yaml file containing the definition of the
                        factory plan
        :type  plan:    str
        :param stage:   Stage to be built
        :type  stage:   str
        :raises QalxFactoryBuildError: - if there is a problem with the build
        """
        self._msg_user(msg="Starting Factory Build")
        self.stage = stage
        self.pack(plan=plan, stage=self.stage, delete=False)

        # Add the key value tag pairs from the factory plan to the session
        for key, val in self._plan_tags.items():
            self.session.tags.add(name=key, value=val)

        def _delete_resources(_delete_stacks=True):
            """
            The build could fail in numerous places.  This private method handles
            consistent messaging and the actual deletion depending on the debug
            value
            """
            if self._plan_debug._demolish_on_failure:
                self._delete_resources(delete_stacks=_delete_stacks)
                # Message the user and raise an exception
                _error = (
                    "Factory Build Failed. No bots on this factory "
                    "have been started. Created resources have been "
                    "deleted. See the logs for more details"
                )
            else:
                # We aren't deleting resources despite the build failure.
                # Explicitly inform the user
                _error = (
                    f"Factory Build Failed. "
                    f"{self._plan_debug._demolish_on_failure_msg}"
                )
            self._msg_user(msg=_error, title="pyqalx", level=logging.ERROR)
            raise QalxFactoryBuildError(_error)

        try:
            self._add_entity(plan=plan)
            self._create_bot_items()
        except QalxAPIResponseError:
            # QalxAPIResponseError: An error occurred when trying to create the
            #                       items.  Unlikely if the user got to this
            #                       point but tidy up should still be attempted
            _delete_resources()
            return
        success = self._create_stacks()

        if success:
            try:
                self._add_factory_bots()
                self._resume_bots()
                self.entity.status = self.DEPLOYED_STATUS
                self.session.factory.save(self.entity)
                self._msg_user(msg="Factory Build Successful")
            except QalxAPIResponseError:
                # QalxAPIResponeError - There was an issue with saving back to
                #                       the API. Unlikely to happen if the user
                #                       has got this far but we
                #                       should still attempt tidy up.
                _delete_resources()
        else:
            # all stacks will have been deleted.  But the other
            # resources need to be tidied up
            _delete_resources(_delete_stacks=False)

    def demolish(self, factory_guid):
        """
        Demolishes a factory

        :param: factory_guid:   The guid of the factory entity to demolish
        """
        self._msg_user(msg="Starting Factory Demolish")
        try:
            self.entity = self.session.factory.get(factory_guid)
            self.entity.status = self.DELETING_STATUS
            self.session.factory.save(self.entity)
            self.breadth_first_terminate()
            self._delete_resources()
        except (ClientError, QalxAPIResponseError) as exc:
            # Catches possible errors and reraises as QalxFactoryDemolishError
            # to give users a consistent API when using programmatically
            # ClientError: An error occurred when trying to delete stacks
            # QalxAPIResponseError: An error occurred when trying to delete
            #                       qalx resources.
            raise QalxFactoryDemolishError(exc)
        self._msg_user(msg="Factory Demolish Successful")



import logging
import os
from io import StringIO

import boto3
import requests
from botocore.exceptions import ClientError
from colorama import Fore
from troposphere import Tags

from pyqalx.core.errors import QalxFactoryError
from pyqalx.factories.sectors.providers.aws.stack import AWSStack
from pyqalx.factories.sectors.providers.provider import BaseProvider


class AWSProvider(BaseProvider):
    """
    A provider subclass that provides interaction with AWS
    """

    _type = "aws"

    FORE = Fore.YELLOW
    REGION_NAME_KEY = "region_name"
    PROFILE_NAME_KEY = "profile_name"
    CREATE_COMPLETE = "CREATE_COMPLETE"
    CREATE_FAILED = "CREATE_FAILED"
    ROLLBACK_IN_PROGRESS = "ROLLBACK_IN_PROGRESS"
    ROLLBACK_FAILED = "ROLLBACK_FAILED"
    ROLLBACK_COMPLETE = "ROLLBACK_COMPLETE"
    DELETE_IN_PROGRESS = "DELETE_IN_PROGRESS"
    DELETE_FAILED = "DELETE_FAILED"
    DELETE_COMPLETE = "DELETE_COMPLETE"

    @staticmethod
    def _identity_document():
        """
        Gets the identity document from the EC2 instance. Certain parameters
        can only be looked up from within the instance (i.e. the default region)
        """
        resp = requests.get(
            "http://169.254.169.254/latest/dynamic/instance-identity/document"
        )
        if resp.ok:
            return resp.json()
        else:
            raise QalxFactoryError(
                f"Error getting " f"identity document: {resp.reason}"
            )

    @staticmethod
    def _get_region():
        """
        Helper for getting the region for remote instances
        """
        return AWSProvider._identity_document()["region"]

    @staticmethod
    def default_profile_name():
        return boto3.session.Session().profile_name

    @staticmethod
    def default_region_name(validate, log):
        session = boto3.session.Session()
        if validate and session.region_name is None:
            # Ensure that when building the local machine has AWS configured
            # properly
            msg = (
                "Must configure the default"
                f" `region_name` for `{AWSProvider._type}`. "
                f"See https://docs.aws.amazon.com/cli/"
                f"latest/reference/configure/#examples"
            )
            AWSProvider._send_message(
                title="all", message=msg, level=logging.ERROR, log=log
            )
            raise QalxFactoryError(msg)
        return session.region_name or AWSProvider._get_region()

    @property
    def ERRORED_STATUSES(self):
        return [
            self.CREATE_FAILED,
            self.ROLLBACK_IN_PROGRESS,
            self.ROLLBACK_FAILED,
            self.ROLLBACK_COMPLETE,
            self.DELETE_IN_PROGRESS,
            self.DELETE_FAILED,
            self.DELETE_COMPLETE,
        ]

    @property
    def CREATED_STATUS(self):
        return self.CREATE_COMPLETE

    @property
    def DELETED_STATUSES(self):
        return [self.DELETE_COMPLETE, self.DELETE_FAILED]

    def _get_client(self, service_name="cloudformation"):
        session = boto3.session.Session(
            profile_name=self.profile_name, region_name=self.region_name
        )
        return session.client(service_name)

    @staticmethod
    def _update_termination_protection(client, enabled, stack_id):
        client.update_termination_protection(
            EnableTerminationProtection=enabled, StackName=stack_id
        )

    @staticmethod
    def _simulate_principal_policy(
        client, session, action_names, resource_arns
    ):
        """
        Helper for simulating the policy for the calling user
        :param session: An instance of ~boto3.session.Session
        :param action_names: A list of action names to simulate
        :param resource_arns: A list of resource arns to simulate
        :return: The response from boto3
        """
        caller_identity = session.client("sts").get_caller_identity()

        resp = client.simulate_principal_policy(
            PolicySourceArn=caller_identity["Arn"],
            ActionNames=action_names,
            ResourceArns=resource_arns,
        )
        return resp

    @staticmethod
    def validate_permissions(profile_name, region_name, log):
        """
        Validates that the calling user has permission to create the
        resources required for the stack
        """
        session = boto3.session.Session(
            profile_name=profile_name, region_name=region_name
        )
        # A list of tuples of <action_names>, <resource_arns> to check
        # the user permission for.
        policies = [
            (
                [
                    "cloudformation:CreateStack",
                    "cloudformation:DeleteStack",
                    "cloudformation:DescribeStacks",
                    "cloudformation:DescribeStackEvents",
                    "cloudformation:UpdateTerminationProtection",
                    "cloudwatch:PutMetricData",
                    "ec2:CreateTags",
                    "ec2:DescribeImages",
                    "ec2:DescribeInstances",
                    "ec2:TerminateInstances",
                    "secretsmanager:CreateSecret",
                    "secretsmanager:TagResource",
                    "ssm:AddTagsToResource",
                    "ssm:PutParameter",
                    "ssm:DeleteParameter",
                    "secretsmanager:DeleteSecret",
                ],
                ["*"],
            ),
            (
                [
                    "iam:AddRoleToInstanceProfile",
                    "iam:CreateInstanceProfile",
                    "iam:CreateRole",
                    "iam:DeleteInstanceProfile",
                    "iam:DeleteRole",
                    "iam:DeleteRolePolicy",
                    "iam:GetInstanceProfile",
                    "iam:GetRole",
                    "iam:GetRolePolicy",
                    "iam:PassRole",
                    "iam:PutRolePolicy",
                    "iam:TagRole",
                    "iam:RemoveRoleFromInstanceProfile",
                ],
                ["arn:aws:iam::*:instance-profile/*", "arn:aws:iam::*:role/*"],
            ),
            (
                ["ec2:RunInstances"],
                [
                    "arn:aws:ec2:*:*:security-group/*",
                    "arn:aws:ec2:*::image/*",
                    "arn:aws:ec2:*:*:instance/*",
                ],
            ),
            (
                [
                    "cloudwatch:EnableAlarmActions",
                    "cloudwatch:DeleteAlarms",
                    "cloudwatch:DisableAlarmActions",
                    "cloudwatch:PutMetricAlarm",
                ],
                ["arn:aws:cloudwatch:*:*:alarm:*"],
            ),
        ]

        results = []
        try:
            client = session.client("iam")
            for action_names, resource_arns in policies:
                resp = AWSProvider._simulate_principal_policy(
                    session=session,
                    client=client,
                    action_names=action_names,
                    resource_arns=resource_arns,
                )
                results += resp["EvaluationResults"]
        except ClientError as exc:
            # ClientError: Problem with calling AWS.  Possibly due to the user
            #              not having permission to `simulate_principal_policy
            AWSProvider._send_message(
                title="all", message=str(exc), level=logging.ERROR, log=log
            )
            raise QalxFactoryError(exc)
        missing_permissions = {}
        for result in results:
            decision = result["EvalDecision"]
            if decision != "allowed":
                missing_permissions[result["EvalActionName"]] = decision

        if missing_permissions:
            formatted_permissions = "\n ".join(
                [
                    f"{action} - {decision},"
                    for action, decision in missing_permissions.items()
                ]
            )

            msg = (
                f"Profile `{profile_name}` on region `{region_name}` is "
                f"missing the following permissions required to create"
                f" factories:\n {formatted_permissions}"
            )
            AWSProvider._send_message(
                title="all", message=msg, level=logging.ERROR, log=log
            )
            raise QalxFactoryError(msg)

    def _create_stack(self, stack):
        """
        Creates the Cloudformation Stack for the given template.  Uploads
        the stack as a file Item to allow us to take advantage of the larger
        file size allowed when using TemplateURL.  This item will be
        automatically deleted when the factory is deleted.
        :param stack:
        :return:
        """
        stack_name = self._stack_name()
        client = self._get_client()
        item = self.factory.session.item.add(
            source=StringIO(stack),
            file_name=self.factory._plan_name,
            meta=self.factory._metadata(self.factory.entity),
        )
        on_failure = (
            "DELETE"
            if self.factory._plan_debug._demolish_on_failure
            else "DO_NOTHING"
        )
        try:
            resp = client.create_stack(
                StackName=stack_name,
                TemplateURL=item.file.url,
                OnFailure=on_failure,
                EnableTerminationProtection=True,
                Capabilities=["CAPABILITY_NAMED_IAM"],
                Tags=Tags(*AWSStack(self)._resource_tags(stack_name)).to_dict(),
            )
        except ClientError as exc:
            # ClientError: There was some sort of issue when
            #              we tried to create the stack.  Record this error and
            #              the manager will ensure that any previous stack
            #              gets deleted.
            self._msg_user(str(exc), level=logging.ERROR)
            return
        self.stack_id = resp["StackId"]
        return self.stack_id

    def _describe_stack(self):
        client = self._get_client()
        try:
            resp = client.describe_stacks(StackName=self.stack_id)
        except ClientError:  # pragma: no cover
            # ClientError: The stack can't be found at all.
            #              This may be because it has been deleted
            return None
        return resp["Stacks"][0]["StackStatus"]

    def _describe_images(self, image_ids):
        """
        Gets the images for the given image_ids so that we can determine
        the platform of the image
        """
        client = self._get_client("ec2")
        try:
            resp = client.describe_images(
                Filters=[{"Name": "image-id", "Values": image_ids}]
            )["Images"]
        except ClientError as exc:
            # ClientError: Some issue communicating with AWS
            self._msg_user(str(exc), level=logging.ERROR)
            return None
        return resp

    @staticmethod
    def _paginated_response(client, list_method, list_key):
        """
        For the given boto3 client and list_method will continue querying AWS
        until `NextToken` is not on the response (meaning we are at the end
        of the pagination)

        :param client: An instance of ~boto3.client
        :param list_method:The list method to use to query AWS
        :param list_key:The key that the response data resides in
        :return:yields an individual response
        """
        # Can't do partial matching, so need them all.
        # Don't catch potential exceptions - this needs to error if
        # there is a problem so that the instance shuts down
        _list_method = getattr(client, list_method)
        parameters = _list_method()
        end = False
        while not end:
            for _p in parameters[list_key]:
                yield _p
            next_token = parameters.get("NextToken")
            if next_token:
                parameters = _list_method(NextToken=next_token)
            else:
                end = True

    def _set_environment_variable(
        self,
        instance_name,
        name,
        client,
        get_method,
        get_kwarg,
        get_key,
        response_key=None,
    ):
        """
        For the given client and get_method will query AWS and return the single
        object (if the name starts with the instance_name).  It will then save
        that as an environment variable
        :param name: The name of the object to lookup
        :param client: An instance of ~boto3.client
        :param get_method: The get method to use to query AWS
        :param get_kwarg: The argument to use when calling the get_method
        :param get_key:The key in the response that the value of the object
                       resides in
        :param response_key: Optional response key in case the returning object
                             is a nested dict
        """
        if name.startswith(instance_name):
            # Qalx parameters will start with the factory guid
            # Don't catch potential exceptions.  This needs to error so that
            # the instance shuts down.
            _variable = getattr(client, get_method)(**{get_kwarg: name})
            # Handle the response being a nested dict
            _variable = _variable.get(response_key, _variable)
            environment_variable_name = self.factory._environment_variable_key(
                name
            )
            os.environ[environment_variable_name] = _variable[get_key]

    def set_environment_variables(self, instance_name):
        """
        Gets ALL the SSM Parameters and Secrets from SecretsManager.
        Iterates through the pages and then sets the ones that start with the
        instance_name as environment variables.  The parameters/secrets are
        created on the Troposphere stack at factory build time.  Uses the
        permissions on the instance role.

        :param instance_name: The name of the instance that was used to create
        the parameters at stack creation.  As parameters are namespace as
        `<<instance_name>_PARAMETER` the name is required to ensure
        the correct parameter is loaded for the instance doing the bootstrapping
        """
        # Will use the role of the instance for permissions
        ssm = self._get_client("ssm")
        for parameter in self._paginated_response(
            client=ssm, list_method="describe_parameters", list_key="Parameters"
        ):
            self._set_environment_variable(
                instance_name=instance_name,
                name=parameter["Name"],
                client=ssm,
                get_method="get_parameter",
                get_kwarg="Name",
                get_key="Value",
                response_key="Parameter",
            )

        secretsmanager = self._get_client("secretsmanager")

        for secret in self._paginated_response(
            client=secretsmanager,
            list_method="list_secrets",
            list_key="SecretList",
        ):
            self._set_environment_variable(
                instance_name=instance_name,
                name=secret["Name"],
                client=secretsmanager,
                get_method="get_secret_value",
                get_kwarg="SecretId",
                get_key="SecretString",
            )

    def _delete_stack(self, stack_id):
        client = self._get_client()
        self._update_termination_protection(
            client=client, stack_id=stack_id, enabled=False
        )
        client.delete_stack(StackName=stack_id)
        return

    def _get_error_reason(self):
        client = self._get_client()
        reasons = []
        events = {"StackEvents": []}
        try:
            events = client.describe_stack_events(StackName=self.stack_id)
        except ClientError:
            # ClientError: Some sort of error (possibly permissions)
            #              when trying to find out what happened
            pass

        for event in events["StackEvents"]:
            if event["ResourceStatus"] == self.CREATE_FAILED:
                reasons.append(event["ResourceStatusReason"])
        return reasons or ["Unknown Error"]

    def _build_stack(self):
        """
        For the given sectors will build a YAML string - one for each
        profile_name/region intersection.  There may be multiple EC2 instances
        per stack if they need to share a profile/region.
        :param sectors: The sectors to create YAML stacks for
        :type sectors:dict
        :return: A dict of profile_name:region_name:yaml_string
        """
        troposphere = AWSStack(self)
        stack_as_yaml = troposphere.generate()
        return stack_as_yaml



from datetime import datetime

import colorama
import logging
from colorama import Fore, Back, Style

APPNAME = "qalx"
APPAUTHOR = "agiletek"


def _msg_user(
    msg,
    log,
    title="pyqalx",
    fore=Fore.WHITE,
    back=Back.BLACK,
    level=logging.DEBUG,
):
    """
    Sends a coloured message to the user and also writes the message
    to the logs
    :param msg: The msg that should be logged
    :param title: The title to print to the console
    :param fore: The foreground colour
    :param back: The background color
    :param log: The log to use to log the message.
    If `NoneType` then don't log
    :param level: The loglevel
    """
    colorama.init()
    bright = Style.BRIGHT
    normal = Style.NORMAL
    if level == logging.ERROR or fore == Fore.RED:
        # Ensure any errors - or those with a RED foreground are displayed
        # using a BRIGHT style
        fore = Fore.RED
        normal = Style.BRIGHT
    now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    title = f"[{now}] {title}"
    title = bright, fore, back, f"{title}:", normal
    message = fore, back, msg, Style.RESET_ALL
    print(*title, *message)

    if log is not None:
        log.log(level=level, msg=msg)



import os

import appdirs

from pyqalx.core.utils import APPNAME, APPAUTHOR

WORKER_LOG_FILE_DIR = os.path.join(appdirs.user_log_dir(APPNAME, APPAUTHOR))
r"""Directory to save worker log files

* Windows: C:\\Users\\<User>\\AppData\\Local\\agiletek\\qalx\\Logs
* Linux: /home/<user>/.cache/qalx/log
* Max ~/Library/Logs/qalx

"""

SAVE_INTERMITTENT = False
"""Save the entity after every processing step?"""

SKIP_FINAL_SAVE = False
"""Skip saving the entity back to qalx after post-processing."""

KILL_AFTER = None
"""The number of attempts to get a message from an empty queue before exiting.
When set to None the worker will keep asking forever."""

Q_WAITTIMESECONDS = 20
"""The minimum time in seconds to wait before asking for a message from an
empty queue"""

Q_MSGBATCHSIZE = 1
"""The number of messages to get in each request"""

Q_WAITTIMEMAXSPREADING = 12
"""The maximum spreading factor for a wait.
The maximum wait time will be ``Q_WAITTIMEMAXSPREADING*Q_WAITTIMESECONDS``."""

MSG_WAITTIMESECONDS = 20
"""The time in seconds the bot will wait for a message to appear in the queue"""

del APPNAME, APPAUTHOR  # These are not config keys



from pyqalx.core.entities.item import Item
from pyqalx.core.entities.entity import QalxQueueableEntity


class Set(QalxQueueableEntity):
    """A set is simply a collection of items.

    They are mapped with keys so that you can retrieve specific items from the
    set later:

    >>> from pyqalx import QalxSession
    >>> qalx = QalxSession()
    >>> dims = qalx.item.add(data={"height":5, "width":5}, meta={"shape":"square"})
    >>> steel = qalx.item.add(source="path/to/316_datasheet.pdf",
    ...                       data={"rho":8000, "E":193e9},
    ...                       meta={"library":"materials",
    ...                             "family":"steel", "grade":"316"})
    >>> steel_square_set = qalx.set.add(items={"shape":dims, "material":steel},
    ...                                 meta={"profile":"square_steel"})

    As with items we can then use the ``find_one`` and ``find`` methods to search
    for sets and easily get the item data:.
    You can use attribute style getters/setters or key style getters/setters

    >>> from pyqalx import QalxSession
    >>> qalx = QalxSession()
    >>> my_set = qalx.set.find_one(query={"metadata.data.profile": "square_steel"})
    >>> # Key style
    >>> youngs_mod = my_set.get_item_data("material")['E']
    >>> # Attribute style
    >>> younds_mod = my_set.get_item_data("material").E

    .. note::
        Sets store a **reference to an item** rather than a copy of the item.
        Changes to the item will be reflected in all sets containing that item.

    """

    entity_type = "set"
    child_entity_class = Item

    def get_item_data(self, item_key):
        """
        helper method to get data from an item in the set

        :param item_key: key on the set to add the item as
        :type item_key: str
        :return: dict or None if key is missing
        """
        if self.items_.get(item_key):
            return self.items_.get(item_key)["data"]
        else:
            return None

    @property
    def items_(self):
        """
        Helper to allow easier lookup for items.  A user can continue to do
        `myset['items']` should they choose - but using this property leads to
        neater code.

        Usage: `myset.items_`

        :return: dict
        :raises: KeyError if `items` key doesn't exist on Set
        """
        # We have to use a property rather than an instance attribute even though
        # it's slower because using an instance attribute would prevent the
        # user from doing `Set({})`.
        return self["items"]



import logging
import os

import appdirs

from ...core.utils import APPNAME, APPAUTHOR

TOKEN = None
"""API token"""

LOGGING_LEVEL = logging._levelToName[logging.ERROR]
"""Level that we should log at by default. """

LOG_FILE_PATH = os.path.join(
    appdirs.user_log_dir(APPNAME, APPAUTHOR), ".qalx.log"
)
r"""The path where log files will be stored.

* Windows: C:\\Users\\<User>\\AppData\\Local\\agiletek\\qalx\\Logs\\.qalx.log
* Linux: /home/<user>/.cache/qalx/log/.qalx.log
* Max ~/Library/Logs/qalx/.qalx.log

"""

MSG_BLACKOUTSECONDS = 30
"""
After this time in seconds the message will be returned to the queue if
it has not been deleted.

.. note::
    Workers automatically issue a heartbeat to the queue while the job is being
    processed to keep extending the blackout during processing
"""

MAX_RETRY_500 = 2
"""How many times should pyqalx retry the request when receiving a 5XX error
back from the API?"""

FACTORY_PACK_DIR = appdirs.user_data_dir(APPNAME, APPAUTHOR)
r"""
The directory where factory bot code will be packed and deployed from.
Defaults to the following:

* Windows: C:\\Users\\<User>\\AppData\\Local\\agiletek\\qalx
* Linux: ~/.local/share/qalx
* Mac: ~/Library/Application Support/qalx
"""

DOWNLOAD_DIR = appdirs.user_data_dir(APPNAME, APPAUTHOR)
r"""
The directory where files from qalx will be downloaded by default.
Defaults to the following:

* Windows: C:\\Users\\<User>\\AppData\\Local\\agiletek\\qalx
* Linux: ~/.local/share/qalx
* Mac: ~/Library/Application Support/qalx
"""

REGION = "eu-west-2"
r"""
The region that will be used when connecting to databases or other remote
services.  Originally set via `qalx configure`.  Change this to another valid
region to use services in that region.
This is only a valid option for User configs - not for Bot configs
"""

del APPNAME, APPAUTHOR  # These are not config keys



import os
import types
from itertools import zip_longest
from time import sleep

import requests
from addict import Dict

from pyqalx.core.encryption import QalxEncrypt

from pyqalx.core.errors import (
    QalxAPIResponseError,
    QalxError,
    QalxQueueError,
    QalxFileRetrievalError,
)


class QalxAddict(Dict):
    """
    A qalx subclass of addict.Dict.  Enables us to have entities that operate
    like normal dicts, but allow dot notation for all the keys.  Ensures that
    if something is passed to it that is already an instance of QalxEntity
    that we don't mutate the type to a QalxAddict instance
    """

    def __getattr__(self, item):
        """
        By default `addict.Dict` will lookup the key when doing attribute access.
        If the key doesn't exist then a KeyError will be raised due to the
        `__missing__` method.  However, this breaks calls to `hasattr()` as that
        catches the normal `AttributeError` that will be raised.  Therefore,
        handle a key not existing by catching the KeyError and reraising as the
        more expected type
        """
        try:
            return super(QalxAddict, self).__getattr__(item)
        except KeyError:
            raise AttributeError(
                "%r object has no attribute %r"
                % (self.__class__.__name__, item)
            )

    def __missing__(self, key):
        """
        https://github.com/mewwts/addict#default-values
        Ensures that if a key/attribute is accessed that is missing that a
        KeyError is raised
        """
        raise KeyError(key)

    @classmethod
    def _hook(cls, item):
        """
        Ensure that if the `item` is an instance of QalxEntity (a dict), that
        we don't mutate the type to be an addict.Dict class.  We want to
        preserve the type and the QalxEntity instance has already gone through
        addict.Dict to allow dot notation lookup
        """
        if not isinstance(item, QalxEntity):
            item = super(QalxAddict, cls)._hook(item)
        return item


class BaseQalxEntity(dict):
    """
    The base class for all QalxEntity instances.  Operates like a normal `dict`
    but utilises `addict.Dict` to allow a user to do dot notation lookup of
    all keys recursively.
    """

    def __init__(self, *args, **kwargs):
        # If the `_region` was specified then this entity was loaded by an
        # instance of QalxAdapter.  Set the `_region` as an instance attribute
        # to be used when interfacing with remote services
        self._region = kwargs.pop("_region", None)

        super(BaseQalxEntity, self).__init__(*args)

        for key, value in self.items():
            # Ensure that `__setitem__` is called for all top level keys.  This
            # will enable dot notation lookup to work.
            self[key] = value

    @staticmethod
    def _addict(value):
        """
        Handles turning the value into a QalxAddict subclass. Will only attempt
        conversion if value is an instance of dict and is not already an instance
        of QalxEntity, or if it is a list/tuple in which case it iterates it
        and attempts conversion on all elements that are a dict and not an
        instance of QalxEntity in which case it will already have dot notation
        available
        """

        def _do_conversion(_value):
            return isinstance(_value, dict) and not isinstance(
                _value, QalxEntity
            )

        if _do_conversion(value):
            value = QalxAddict(value)
        elif isinstance(value, (list, tuple)):
            value = [QalxAddict(x) if _do_conversion(x) else x for x in value]
        return value

    def _reserved_keys(self):
        """
        A list of reserved keys for this entity.  Keys that are listed in here
        will never be assigned to the dict value when setting the attribute.
        i.e. if `_reserved_keys` returns ["_broker_client"] then doing
        `myentity._broker_client = None` would set the attribute of
        `_broker_client` but NOT the dict key.  Only applies for top level
        dict keys
        """
        return ["_region"]

    def __setitem__(self, key, value):
        """
        Ensure that the appropriate attribute is set for `key`.
        Takes advantage of dict mutability so that if a user does
         `myentity.data.nested.value = 5` it will update the attribute
        and the dict key.  The same goes for if the user does
        `myentity['data']['nested']['value'] = 5`
        """
        value = self._addict(value)
        super(BaseQalxEntity, self).__setitem__(key, value)
        # Now that the dict has been updated with an addict value, ensure
        # that the attribute is kept in sync.  Ensure that we don't override
        # any builtins.  This means a user will have to do `entity['items']`
        # if they want to access `items` - which keeps `entity.items()` working
        # like the baseclass
        if not isinstance(
            getattr(self, key, None),
            (types.BuiltinMethodType, types.MethodType),
        ):
            setattr(self, key, value)

    def __setattr__(self, key, value):
        """
        Ensures that if a user updates a root level attribute
        (i.e. myentity.meta = {"new": "meta"}) that the dict key is updated
        to keep the dict and attribute values in sync.  Ignores specific
        reserved keys which are specifically set on the instance.
        """
        super(BaseQalxEntity, self).__setattr__(key, value)
        if key not in self._reserved_keys():
            sentinel = object()
            if self.get(key, sentinel) != value:
                self[key] = value

    def to_dict(self):
        base = {}
        for key, value in self.items():
            base[key] = value.to_dict()
        return base


class QalxListEntity(BaseQalxEntity):
    """
    Simple wrapper around a pyqalxapi_dict so we can keep extra keys
    on the API list response.  Instantiates each entity in `data` to the
    correct QalxEntity subclass.

    :param pyqalxapi_list_response_dict: A dict response from a REST API list endpoint
    :type pyqalxapi_list_response_dict: dict
    """

    _data_key = "data"

    def __init__(self, pyqalxapi_list_response_dict, **kwargs):
        self.child = kwargs.pop("child")
        self._key_file = kwargs.pop("keyfile", None)
        super(QalxListEntity, self).__init__(
            pyqalxapi_list_response_dict, **kwargs
        )

        if (
            self._data_key not in pyqalxapi_list_response_dict
            or not isinstance(
                pyqalxapi_list_response_dict[self._data_key], list
            )
        ):
            raise QalxAPIResponseError(
                "Expected `{0}` key in "
                "`pyqalxapi_list_response_dict` and for"
                " it to be a list".format(self._data_key)
            )

    def _reserved_keys(self):
        reserved_keys = super(QalxListEntity, self)._reserved_keys()
        reserved_keys += ["_key_file", "child"]
        return reserved_keys

    def __setitem__(self, key, value):
        """
        Ensures that if `key == self._data_key` that all the entities on the
        data key are instantiated to be `self.child` entities.
        """
        if key == self._data_key:
            kwargs = {}
            if self._key_file is not None:
                kwargs = {"keyfile": self._key_file}
            # Ensure the region is set for all children
            kwargs["_region"] = self._region
            value = [self.child(x, **kwargs) for x in value]
        super(QalxListEntity, self).__setitem__(key, value)


class AggregationResult(QalxListEntity):
    """
    Class for aggregation result responses from the API

    :param pyqalxapi_list_response_dict: A dict response from a REST API list endpoint
    :type pyqalxapi_list_response_dict: dict
    """


class QalxEntity(BaseQalxEntity):
    """Base class for a response from the qalx API

    QalxEntity children need to be populated with either a
    `requests.models.Response` which is the type returned by the methods
    on :class:`~pyqalx.transport.api.PyQalxAPI` or with a `dict`.

    Entities can behave either like a dict or attribute lookups can be used
    as getters/setters

    >>> class AnEntity(QalxEntity):
    ...     pass
    >>> c = AnEntity({"guid":"123456789", "info":{"some":"info"}})
    >>> # dict style lookups
    >>> c['guid']
    '123456789'
    >>> # attribute style lookups
    >>> c.guid
    '123456789'


    :param pyqalxapi_dict: a 'dict' representing a qalx entity object to
        populate the entity
    :type pyqalxapi_dict: dict
    """

    entity_type: str

    @classmethod
    def _chunks(cls, _iterable, chunk_size, fillvalue=None):
        """
        Collect data into fixed-length chunks or blocks"
        # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
        Taken from the itertools documentation
        """
        args = [iter(_iterable)] * chunk_size
        return zip_longest(fillvalue=fillvalue, *args)

    @classmethod
    def pluralise(cls):
        """
        Pluralises the entity type
        """
        vowels = {"a", "e", "i", "o", "u", "y"}
        singular = cls.entity_type
        plural = singular + "s"
        if len(singular) > 1 and singular.endswith("y"):
            if singular[-2] not in vowels:
                plural = singular[:-1] + "ies"
        return plural


class QalxFileEntity(QalxEntity):
    _file_bytes = None
    _file_key = "file"

    def __init__(self, *args, **kwargs):
        """
        QalxFileEntity subclasses can be provided with a `keyfile` kwarg.
        This gets set on the instance of the entity and will allow for seamless
        decryption of files
        """
        self._key_file = kwargs.pop("keyfile", None)
        super(QalxFileEntity, self).__init__(*args, **kwargs)

    def _reserved_keys(self):
        reserved_keys = super(QalxFileEntity, self)._reserved_keys()
        reserved_keys += ["_key_file", "_file_bytes"]
        return reserved_keys

    def read_file(self, retry=True):
        """
        If this Item contains a file, will read the file data and cache it
        against the Item.

        :param retry:   Should the download retry if there is a problem
                        getting the file from S3?  By default, the download
                        will retry once
        :type retry: bool
        :return:    The content of the URL as a bytes object.  Accessible from
                    the `_file_bytes` attribute
        :raises QalxError:
        """
        if not self.get(self._file_key):
            raise QalxError("Item doesn't have file data.")
        else:
            response = requests.get(url=self[self._file_key]["url"])
            if response.ok:
                # File retrieved Ok
                file_bytes = response.content
                if (
                    self[self._file_key].get("keyfile") is not None
                    and self._key_file is not None
                ):
                    qalx_encrypt = QalxEncrypt(self._key_file)
                    file_bytes = qalx_encrypt.decrypt(file_bytes)
                self._file_bytes = file_bytes
                return self._file_bytes
            else:
                # Problem getting the file - possibly S3 has had a minor
                # hiccough. This is probably only temporary so wait a couple of
                # seconds and try again (ensuring that we raise an exception if
                # the download fails a second time)
                if retry:
                    sleep(2)
                    return self.read_file(retry=False)
                # If we're not retrying then raise an error with the error message
                raise QalxFileRetrievalError(
                    "Error with file retrieval: \n\n" + response.text
                )

    def save_file_to_disk(self, filepath, filename=None):
        """
        If this Item contains a file, will read the file from the URL (or from
        the cached bytes on the instance) and save the file to disk.  Provide
        an optional `filename` argument if you don't want to use the same
        filename as the one stored on the Item

        :param filepath: The path where this file should be saved
        :type filepath: str
        :param filename: The optional name of this file. Defaults to the name
            of the file on the instance
        :type filename: str
        :param key_file:    The path to the encryption key to be used for
                            decrypting the file data. If it is not provided
                            and the data was encrypted, the data will be read
                            in the encrypted format
        :type key_file:     str
        :returns: Path to file on disk
        :raises QalxError:
        """
        if self._file_bytes is None:
            self.read_file()
        if filename is None:
            filename = self[self._file_key]["name"]
        _filepath = os.path.join(filepath, filename)
        with open(_filepath, "wb") as f:
            f.write(self._file_bytes)
        return _filepath


class QalxQueueableEntity(QalxEntity):
    """
    A mixin which allows an entity to be sumitted to queue
    """

    @classmethod
    def add_to_queue(cls, payload, queue, children=False, **message_kwargs):
        """
        Submits an entity, entity guid or list of entities/entity guids
        to the given queue

        Example usage for an Item:

        >>> Item.add_to_queue(item, queue)
        >>> Item.add_to_queue(item.guid, queue)
        >>> Item.add_to_queue([item, item.guid], queue)

        :param payload: subclassed instance of
                        :class:`~pyqalx.core.entities.entity.QalxEntity`,
                        a guid, or a list containing a combination of both
        :param queue: An instance of :class:`~pyqalx.core.entities.queue.Queue`
        :type queue: ~pyqalx.core.entities.queue.Queue
        :param children: Whether we are submitting the child entities to the
                         queue rather than the given entity
        :type children: bool
        """
        entity_type = cls.entity_type
        if children:
            #  Get the children off this entity and submit them to the queue
            if not hasattr(cls, "child_entity_class"):
                raise QalxQueueError(
                    f"`{cls}` is not an entity that " f"has children."
                )
            if not isinstance(payload, cls):
                raise QalxQueueError(
                    f"Expected payload of type `{cls}"
                    f"` when calling `{cls.__name__}.add_to_queue` "
                    f"with `children=True`. "
                    f"Got `{type(payload)}`."
                )
            # SQS needs the guid as a string - not UUID
            message_kwargs[f"parent_{entity_type}_guid"] = str(payload["guid"])
            entity_type = cls.child_entity_class.entity_type
            payload = list(payload[cls.child_entity_class.pluralise()].values())
        return queue._send_message(
            payload=payload, entity_type=entity_type, **message_kwargs
        )



"""
pyqalx.core.errors defines QalxError exception and a load of children.

If pyqalx is going raise an error you know about then use one of these or
create a new one.
"""


class QalxError(Exception):
    """Base qalx error. Take responsibility!"""


class QalxAuthError(QalxError):
    """
    qalx did not find a way to authenticate or the authentication didn't work
    """


class QalxNoGUIDError(QalxError):
    """A QalxEntity without a guid is like a dog without a bone."""


class QalxNoInfoError(QalxError):
    """A QalxEntity without info is like a dog without a bone."""


class QalxConfigProfileNotFound(QalxError):
    """The profile wasn't in the file or the file wasn't properly formed."""


class QalxConfigFileNotFound(QalxError):
    """
    There should be a file in the users `home` directory (either a .bots or
    .qalx)
    """


class QalxQueueError(QalxError):
    """There wasn't the correct information to connect to the remote queue."""


class QalxBotInitialisationFailed(QalxError):
    """The bot initialisation function returned something falsey."""


class QalxEntityTypeNotFound(QalxError):
    """We couldn't find the entity type you were looking for."""


class QalxEntityNotFound(QalxError):
    """We couldn't find the entity you were looking for."""


class QalxMultipleEntityReturned(QalxError):
    """We found more than one entity, but you just wanted the one hey?"""


class QalxConfigError(QalxError):
    """Something about an attempted load of config didn't work"""


class QalxAPIResponseError(QalxError):
    """There was a problem with some kind of API request."""


class QalxEntityUnchanged(QalxError):
    """
    Saved something which hadn't actually been changed when we thought it had.
    """


class QalxInvalidSession(QalxError):
    """
    The qalx_session argument passed to an adapter isn't a valid QalxSession
    instance
    """


class QalxNoEntity(QalxError):
    """
    A user tried to access the `entity` attribute on QalxAdapter when they
    hadn't set an entity
    """


class QalxIncorrectEntityType(QalxError):
    """
    The entity on the QalxAdapter is of a different type to the type of adapter
    """


class QalxFileError(QalxError):
    """
    There is something wrong with the file details passed to the QalxAdapter
    """


class QalxAlreadyRegistered(QalxError):
    """The entity is already registered with the session"""


class QalxCannotUnregister(QalxError):
    """The entity cannot be unregistered"""


class QalxRegistrationClassNotFound(QalxError):
    """The registration class for registering a custom class was not found"""


class QalxInvalidBlueprintError(QalxError):
    """The blueprint is not valid."""


class QalxInvalidTagError(QalxError):
    """The user does not have access to write to the specific tags"""


class QalxStepFunctionNotDefined(QalxError):
    """A specific Bot step function has not been defined by the user"""


class QalxFileRetrievalError(QalxError):
    """Error when retrieving a file from qalx"""


class QalxFactoryError(QalxError):
    """Generic error with a Qalx factory"""


class QalxFactoryValidationError(QalxFactoryError):
    """Error when validating a factory"""


class QalxFactoryPackError(QalxFactoryError):
    """Error when packing a factory"""


class QalxFactoryBuildError(QalxFactoryError):
    """Error when building a factory"""


class QalxFactoryDemolishError(QalxFactoryError):
    """Error when demolishing a factory"""



import concurrent
from io import StringIO, BytesIO
from typing import Union

from pyqalx.core.adapters.adapter import (
    QalxValidateBlueprintAdapter,
    QalxFileAdapter,
)
from pyqalx.core.entities import Item
from pyqalx.core.entities.item import ItemAddManyEntity
from pyqalx.core.errors import QalxError, QalxAPIResponseError


class QalxItem(QalxValidateBlueprintAdapter, QalxFileAdapter):
    """
    Provides an interface for accessing the API for
    :class:`~pyqalx.core.entities.item.Item` entities
    """

    _entity_class = Item

    def _validate(
        self,
        data=None,
        source=None,
        file_name="",
        meta=None,
        blueprint_name=None,
        as_file=False,
        **kwargs,
    ):
        if data is None:
            data = {}
        return super(QalxItem, self)._validate(
            data=data,
            source=source,
            file_name=file_name,
            meta=meta,
            blueprint_name=blueprint_name,
            as_file=as_file,
            **kwargs,
        )

    def add(
        self,
        data: dict = None,
        source: Union[str, StringIO, BytesIO] = None,
        file_name: str = "",
        meta: dict = None,
        blueprint_name: str = None,
        upload: bool = True,
        encrypt: bool = True,
        as_file: bool = False,
        **kwargs,
    ):
        """
        Adds an `Item` instance that can contain either
        data (as a dict), a file or both.

        :param data: Optional data to store against this Item
        :param source: file path or instance of StringIO or BytesIO
         (https://docs.python.org/3/library/io.html)
        :param file_name: input file name. Optional if a file path is given
         for source.  Required if an `io` object
        :param meta: A dictionary of metadata to store
        :param blueprint_name: An optional blueprint name to use to validate
         this item against an existing Blueprint
        :param upload: Whether the file should be automatically uploaded or not
        :param encrypt: Whether the file should be automatically encrypted if
         the KEYFILE is present
        :param as_file: Whether the `data` dict should be stored in S3 rather
         than in Mongo

        :return: :class:`~pyqalx.core.entities.item.Item` instance
        """
        return super(QalxItem, self).add(
            data=data,
            source=source,
            file_name=file_name,
            meta=meta,
            blueprint_name=blueprint_name,
            upload=upload,
            encrypt=encrypt,
            as_file=as_file,
            **kwargs,
        )

    def save(
        self,
        entity,
        source=None,
        file_name="",
        blueprint_name=None,
        upload=True,
        encrypt=True,
        as_file=False,
        **kwargs,
    ):
        """
        Saves an updated existing `Item` instance.
        To remove a file from an `Item` update the entity so that
        `entity['file'] = {}`

        :param entity: The entity that we are saving
        :type entity: An instance of `Item`
        :param source: file path or instance of StringIO or
            BytesIO (https://docs.python.org/3/library/io.html)
        :type source: str or `io` object
        :param file_name: input file name. Optional if a file path is given
            for source.  Required if an `io` object
        :type file_name: str
        :param meta: A dictionary of metadata to store
        :type meta: dict
        :param blueprint_name: An optional blueprint name to use if you want
                               to validate this item against an existing
                               Blueprint
        :type blueprint_name: str
        :param upload: Whether the file should be automatically uploaded or not
        :type upload: bool
        :param encrypt: Whether the file should be automatically encrypted if
         the KEYFILE is present
        :type encrypt: bool
        :param as_file: Whether the `data` dict should be stored in S3 rather
         than in Mongo
        :return: An updated `Item` instance
        """
        return super(QalxItem, self).save(
            entity=entity,
            source=source,
            file_name=file_name,
            blueprint_name=blueprint_name,
            upload=upload,
            encrypt=encrypt,
            as_file=as_file,
            **kwargs,
        )

    def add_many(self, items, encrypt=True, as_file=False, **kwargs):
        """
        Adds multiple items at once.

        :param items: A list of item data that you want to create.
                      This can be a mixture of non file items and file items.
                      The keys of each item in the list should be the same
                      as if a single item were being created via the `add` method
        :type items: list
        :param encrypt: Whether the file should be automatically encrypted if
                        the KEYFILE is present.  All files are encrypted or
                        not encrypted based on this value
        :type encrypt: bool
        :return:
        """
        if not isinstance(items, list):
            raise QalxError("`items` argument must be a list")
        headers = {"qalx-bulk": "true"}
        validated_items = []

        for item in items:
            # Ensure all the items are valid.  They should be in the same
            # format as if a user is adding a single item.  This also handles
            # checking the blueprint for each item
            # tags should be added on a per item basis
            self._add_tags(item)
            validated_items.append(self._validate(encrypt=encrypt, **item))

        resp = self.add(
            items=validated_items,
            headers=headers,
            _is_validated=True,
            # Don't upload as we need to handle
            # the file upload manually
            upload=False,
            parent_class=ItemAddManyEntity,
            as_file=as_file,
            **kwargs,
        )

        # We have all the data to create the files, but the indexes have
        # been split up compared to the response.  Therefore, inject blank
        # values into the response where there are errors.  This means that
        # the structure will match the original `items` that the user specified
        # (as the API preserves order when bulk creating data)
        ERRORED_KEY = None
        for _ in resp.errors.keys():
            resp.items_.insert(int(_), ERRORED_KEY)
        uploaded_files = {}

        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
            # Allows for partial success (i.e. some files uploaded successfully,
            # others did not)
            for index, item in enumerate(items):
                # Iterate the original `items` provided by the user
                created_item = resp.items_[index]
                if item.get("source") and created_item is not None:
                    # There was a file on the original item and it was also
                    # successfully created by the API.  Upload the file as normal
                    uploaded_files[
                        executor.submit(
                            self._upload_file,
                            entity=resp.items_[index],
                            source=item["source"],
                        )
                    ] = index
            for future in concurrent.futures.as_completed(uploaded_files):
                # Handles updating the error response if any files failed to
                # upload
                index = uploaded_files[future]
                try:
                    future.result()
                except QalxAPIResponseError as exc:
                    # There was an error uploading this single file.  Update
                    # the response dict accordingly.
                    resp.items_.pop(index)
                    resp.items_created = resp.items_created - 1
                    resp.items_errored = resp.items_errored + 1
                    resp.errors[index] = exc

        # Finally, remove all the `None` values from the items as these were
        # errors that we are done with.  This needs to remain as a dict assignment
        # so that the `items_` property continue to work properly
        resp.items = list(filter(lambda x: x is not ERRORED_KEY, resp.items_))
        if resp.errors:
            for index, error in resp.errors.items():
                self.session.log.error(
                    f"Error bulk creating at index `{index}`. "
                    f"Error: `{error}`. "
                    f"Original Data: `{items[int(index)]}`"
                )
        return resp

    def find(
        self,
        query=None,
        sort=None,
        skip=0,
        limit=25,
        many=True,
        include_session_tags=True,
        **kwargs,
    ):
        # extended to create better autodocs
        return super(QalxItem, self).find(
            query=query,
            sort=sort,
            skip=skip,
            limit=limit,
            many=many,
            include_session_tags=include_session_tags,
            **kwargs,
        )

    def find_one(self, query, **kwargs):
        # extended to create better autodocs
        return super(QalxItem, self).find_one(query=query, **kwargs)

    def reload(self, entity, *args, **kwargs):
        # extended to create better autodocs
        return super(QalxItem, self).reload(entity=entity, **kwargs)

    def aggregate(self, aggregate, **kwargs):
        # extended to create better autodocs
        return super(QalxItem, self).aggregate(aggregate=aggregate, **kwargs)

    def archive(self, entity, *args, **kwargs):
        # extended to create better autodocs
        return super(QalxItem, self).archive(entity, *args, **kwargs)

    def delete(self, entity, *args, **kwargs):
        # extended to create better autodocs
        return super(QalxItem, self).delete(entity, *args, **kwargs)



from pyqalx.core.entities.entity import QalxFileEntity


class Factory(QalxFileEntity):
    """
    A factory is an entity that allows for the interaction of bots that can
    execute code from different locations

    Class attributes
        entity_type:    The qalx type is factory
    """

    entity_type = "factory"
    _file_key = "plan"



import logging

import click

from pyqalx import QalxSession
from pyqalx.core.errors import (
    QalxFactoryValidationError,
    QalxFactoryPackError,
    QalxFactoryBuildError,
    QalxFactoryDemolishError,
)
from pyqalx.factories.factory import Factory
from pyqalx.factories.sectors.providers import AWSProvider

from .. import cli_types
from ..cli.qalx import qalx
from ..utils import DemolishFactoryTabulation, FactoryTabulation


@qalx.command("factory-validate")
@click.option(
    "--plan",
    type=click.Path(exists=True),
    help="The path to the file that contains the factory plan definition",
    required=True,
)
@click.option(
    "-aws",
    "--aws-profile",
    help="The AWS profile name to use for remote stacks",
    default=AWSProvider.default_profile_name(),
)
@click.pass_context
def validate(ctx, plan, aws_profile):
    """
    Validate a factory
    """
    session = QalxSession(profile_name=ctx.obj["USER_PROFILE"])
    factory = Factory(
        session,
        user_profile=ctx.obj["USER_PROFILE"],
        bot_profile=ctx.obj["BOT_PROFILE"],
        aws_profile=aws_profile,
    )

    try:
        factory.validate(plan=plan)
    except QalxFactoryValidationError as exc:
        # QalxFactoryValidationError: An error occurred while validating
        factory._msg_user(title="validation", msg=str(exc), level=logging.ERROR)
        factory._msg_user(msg="Factory Validation Failed")


@qalx.command("factory-pack")
@click.option(
    "--stage", type=cli_types.QALX_FACTORY_STAGE, help="The stage to pack"
)
@click.option(
    "--plan",
    type=click.Path(exists=True),
    help="The path to the file that contains the factory plan definition",
    required=True,
    is_eager=True,
)
@click.option(
    "--no-delete",
    is_flag=True,
    help="If provided will *not* delete the packed bots when this command"
    " completes. By default, the packed bots will be deleted.",
)
@click.option(
    "-aws",
    "--aws-profile",
    help="The AWS profile name to use for remote stacks",
    default=AWSProvider.default_profile_name(),
)
@click.pass_context
def pack(ctx, plan, stage, aws_profile, no_delete):
    """
    Pack a factory
    """
    session = QalxSession(profile_name=ctx.obj["USER_PROFILE"])
    factory = Factory(
        session,
        user_profile=ctx.obj["USER_PROFILE"],
        bot_profile=ctx.obj["BOT_PROFILE"],
        aws_profile=aws_profile,
    )
    try:
        factory.pack(plan=plan, stage=stage, delete=not no_delete)
    except (QalxFactoryValidationError, QalxFactoryPackError) as exc:
        # QalxFactoryValidationError: An error occurred while validating
        # QalxFactoryPackError: An error occurred while packing
        factory._msg_user(title="pack", msg=str(exc), level=logging.ERROR)
        factory._msg_user(msg="Factory Pack Failed")


@qalx.command("factory-build")
@click.option(
    "--stage",
    type=cli_types.QALX_FACTORY_STAGE,
    help="The stage to build",
    required=True,
)
@click.option(
    "--plan",
    type=click.Path(exists=True),
    help="The path to the file that contains the factory plan definition",
    required=True,
    is_eager=True,
)
@click.option(
    "-aws",
    "--aws-profile",
    help="The AWS profile name to use for remote stacks",
    default=AWSProvider.default_profile_name(),
)
@click.pass_context
def build(ctx, plan, stage, aws_profile):
    """
    Build a factory
    """
    session = QalxSession(profile_name=ctx.obj["USER_PROFILE"])
    factory = Factory(
        session,
        user_profile=ctx.obj["USER_PROFILE"],
        bot_profile=ctx.obj["BOT_PROFILE"],
        aws_profile=aws_profile,
    )
    try:
        factory.build(plan=plan, stage=stage)
    except (
        QalxFactoryValidationError,
        QalxFactoryPackError,
        QalxFactoryBuildError,
    ) as exc:
        # QalxFactoryValidationError: An error occurred while validating
        # QalxFactoryPackError: An error occurred while packing
        # QalxFactoryBuildError: There was an error with building. Most of
        #                        these are handled internally by build as
        #                        resources need to be tidied up
        factory._msg_user(title="build", msg=str(exc), level=logging.ERROR)
        factory._msg_user(msg="Factory Build Failed")


@qalx.command("factory-demolish")
@click.option(
    "--name", type=click.STRING, help="The name of the factory", required=True
)
@click.pass_context
def demolish(ctx, name):
    """
    Demolish a factory. If a factory to demolish is identified, the user is
    asked for confirmation before the demolish method is called
    """
    session = QalxSession(profile_name=ctx.obj["USER_PROFILE"])
    factory = Factory(session)
    factory_tabulation = DemolishFactoryTabulation(session, name)
    factory_to_demolish = factory_tabulation.get_entity_or_display()

    if factory_to_demolish:
        if click.confirm(f"Are you sure you want to demolish factory: {name}?"):
            try:
                factory.demolish(factory_to_demolish.guid)
            except QalxFactoryDemolishError as exc:
                # QalxFactoryDemolishError: There was an error with demolishing
                factory._msg_user(
                    title="demolish", msg=str(exc), level=logging.ERROR
                )
                factory._msg_user(msg="Factory Demolish Failed")


@qalx.command("factory-info")
@click.option(
    "--name", type=click.STRING, help="The name of the factory", required=True
)
@click.pass_context
def info(ctx, name):
    """
    Get info about a factory
    """
    session = QalxSession(profile_name=ctx.obj["USER_PROFILE"])
    factory_tabulation = FactoryTabulation(session, name, tabulate_single=True)
    factory_tabulation.get_entity_or_display()



import importlib
from collections import defaultdict

from pyqalx.core.entities import QalxEntity
from pyqalx.core.errors import (
    QalxAlreadyRegistered,
    QalxIncorrectEntityType,
    QalxCannotUnregister,
    QalxRegistrationClassNotFound,
)


class Registry(defaultdict):
    def __init__(self, *args, **kwargs):
        super(Registry, self).__init__(*args, **kwargs)

        # Registry but always be a `dict`
        self.default_factory = dict
        # The classes that we allow to be registered.  Only classes defined
        # here will get registered by the system.
        self.registration_classes = [EntityRegistry]

    def _register_defaults(self):
        """
        Iterates the valid classes that we allow to register and registers
        the defaults that are defined by pyqalx itself
        """
        for registration_class in self.registration_classes:
            register_class = registration_class()

            for default_class in register_class._get_defaults():
                register_class.register(cls=default_class)
            self.update(register_class)

    @property
    def base_class(self):
        """
        The base class is used to prove registry subclass level validation
        on the class being registered
        :raises: NotImplementedError
        """
        raise NotImplementedError(
            "`base_class` must be defined on implementing class"
        )

    def _lookup_key(self, cls):
        """
        The key to use to uniquely identify the registered class
        :param cls: The class to register
        :raises: NotImplementedError
        """
        raise NotImplementedError(
            "`lookup_key` must be defined on implementing class"
        )

    @staticmethod
    def _get_package(cls):
        """
        Gets the package the given class is in
        :param cls: The class to get the package for (i.e. `QalxItem`)
        :return: The package in dot notation (i.e. `pyqalx.core.entities`)
        """
        return importlib.import_module(cls.__module__).__package__

    def _get_defaults(self, _cls=None):
        """
        Recursively gets all default subclasses of the registrys `base_class`
        """
        all_subclasses = set()

        def _append_subclasses(_subclasses):
            # Appends the subclass of the given subclasses
            for _subclass in _subclasses:
                if hasattr(_subclass, "entity_type"):
                    # Ignore _subclasses without an entity type.  This allows
                    # the use of interim classes for shared functionality
                    all_subclasses.add(_subclass)
            return all_subclasses

        if _cls is None:
            _cls = self.base_class
        package = self._get_package(self.base_class)

        for subclass in _cls.__subclasses__():
            subclass_package = self._get_package(subclass)

            if subclass_package == package:
                all_subclasses = _append_subclasses([subclass])
                subclasses = self._get_defaults(_cls=subclass)
                all_subclasses = _append_subclasses(subclasses)
        return all_subclasses

    @property
    def _registry_key(self):
        """
        The key to group similar registered classes.
        Uses the right most part of the package name
        :return: Given `pyqalx.core.entities` as a package will return
                 `entities`
        """
        return self._get_package(cls=self.base_class).rsplit(".", 1)[1]

    def _get_registry_class(self, cls):
        """
        Determines the registry class to use based on the given cls. We need
        a specific registry class to perform specific validation
        :param cls: The class to get the registry class for.
        :return: The specific registry subclass
        :raises: QalxRegistrationClassNotFound
        """
        registry_class = None
        for registration_class in self.registration_classes:
            if issubclass(cls, registration_class.base_class):
                registry_class = registration_class()
                break

        if registry_class is None:
            valid_base_classes = [
                str(x.base_class) for x in self.registration_classes
            ]
            msg = (
                f"Could not find registration class for `{cls}`.  Ensure "
                f"that you are registering a class which is a subclass of"
                f' `{", ".join(valid_base_classes)}`'
            )
            raise QalxRegistrationClassNotFound(msg)
        return registry_class

    @staticmethod
    def _validate(cls):
        """
        Hook for providing registry class level validation
        """
        return

    @property
    def registered_classes(self):
        """
        A list of the current registered classes for the `registry_key`
        """
        registered_classes = []
        for _cls in self[self._registry_key].values():
            registered_classes.append(_cls)
        return registered_classes

    def register(self, cls):
        """
        Register the given class to the registry.
        :param cls: The subclass to be registered
        :raises: QalxInvalidSubclass
        :raises: QalxAlreadyRegistered
        """
        RegistrationClass = self
        if isinstance(self, Registry):
            # A `QalxSession` registers via a `Registry()` instance so we
            # need to lookup the specific registry class here to be able to
            # be able to perform validation
            RegistrationClass = self._get_registry_class(cls)
        lookup_key = RegistrationClass._lookup_key(cls)

        # Handle registry level validation
        RegistrationClass._validate(cls)

        if cls in self[RegistrationClass._registry_key].values():
            raise QalxAlreadyRegistered(f"{cls} is already registered")
        # Register the class to the specific `_registry` and `lookup` keys
        self[RegistrationClass._registry_key][lookup_key] = cls

    def unregister(self, cls):
        """
        Unregisters the given adapter_class and reregisters the default
        value
        :param cls: The class you want to unregister
        :raises: QalxCannotUnregister
        """
        RegistrationClass = self._get_registry_class(cls)
        if RegistrationClass._get_package(
            cls
        ) == RegistrationClass._get_package(RegistrationClass.base_class):
            # Prevent deregistration of default adapters
            raise QalxCannotUnregister(
                f"You cannot unregister default class `{cls}`"
            )

        for default_class in RegistrationClass._get_defaults():
            if issubclass(cls, default_class):
                # reregister the default
                self[RegistrationClass._registry_key][
                    RegistrationClass._lookup_key(default_class)
                ] = default_class


class EntityRegistry(Registry):
    base_class = QalxEntity

    def _validate(self, cls):
        entity_type = getattr(cls, "entity_type", None)

        if not entity_type:
            raise QalxIncorrectEntityType(
                f"{cls} must have an attribute `entity_type` set"
            )

    def _lookup_key(self, cls):
        """
        The key to use to uniquely identify the registered class
        :param cls: The class to register
        :returns: cls.entity_type
        """
        return cls.entity_type



import concurrent
import copy
import io
import os
import time
import uuid
from concurrent.futures import ThreadPoolExecutor, as_completed
from http import HTTPStatus

from types import MethodType

import requests
from bson_extra import bson_extra

from pyqalx.config import UserConfig, BotConfig
from pyqalx.core.entities import QalxListEntity
from pyqalx.core.entities.entity import AggregationResult
from pyqalx.core.enums import ValidStates
from pyqalx.core.errors import (
    QalxError,
    QalxNoGUIDError,
    QalxAPIResponseError,
    QalxMultipleEntityReturned,
    QalxEntityNotFound,
    QalxConfigError,
    QalxInvalidSession,
    QalxFileError,
)


class QalxAdapter(object):
    """
    The base class for a QalxAdapter. An adapter is the interface for the
    entity to the rest api.  This allows us to have a consistent interface
    across all entity types.
    """

    # Certain keys shouldn't be attemped to be saved
    _keys_to_exclude = {"info", "guid", "status"}

    def __init__(self, session, *args, **kwargs):
        QalxSession = __import__("pyqalx").QalxSession
        if not isinstance(session, QalxSession):
            raise QalxInvalidSession(
                f"`qalx_session` should be an instance of"
                f" `{QalxSession}`.  "
                f"Got `{type(session)}`"
            )
        self.session = session
        super(QalxAdapter, self).__init__(*args, **kwargs)

    def __getattribute__(self, item):
        """
        Certain methods can only be called by sessions with a UserConfig or a
        BotConfig.  The API will return a 403 Permission Denied
        but this method handles showing the user a more useful error message
        """
        attr = super(QalxAdapter, self).__getattribute__(item)
        if callable(attr) and isinstance(attr, MethodType):
            # This is a method on `QalxBot`.  Check the config
            # is correct for the method we are calling
            def _msg(expected_class, actual_class):
                return (
                    f"Method `{item}` on `{self.__class__}` "
                    f"must be called via a session with a `{expected_class}`"
                    f"instance.  Got `{actual_class}`"
                )

            is_user_config = isinstance(self.session.config, UserConfig)
            is_bot_config = isinstance(self.session.config, BotConfig)

            if (
                item in getattr(self, "_user_only_methods", [])
                and not is_user_config
            ):
                raise QalxConfigError(_msg(UserConfig, BotConfig))
            if (
                item in getattr(self, "_bot_only_methods", [])
                and not is_bot_config
            ):
                raise QalxConfigError(_msg(BotConfig, UserConfig))
        return attr

    @property
    def entity_class(self):
        return self.session._registry["entities"][
            self._entity_class.entity_type
        ]

    @staticmethod
    def raise_api_response_error(data_dict):
        message = (
            "API request error, message:\n\n-vvv-\n\n"
            + "\n".join([f"{k}: {v}" for k, v in data_dict.items()])
            + "\n\n-^^^-"
        )
        raise QalxAPIResponseError(message)

    def _process_api_request(self, method, *args, **kwargs):
        """calls to transport layer

        :param method: http method required
        :param args: args to be passed to endpoint method
        :param kwargs: kwargs to be passed to endpoint method
        :returns: `dict` containing API resource data
        """
        headers = kwargs.pop("headers", {})

        try:
            bson_extra.dumps(kwargs)
        except (TypeError, ValueError, OverflowError):
            raise QalxError(
                "One of the keyword arguments is not JSON serializable"
            )

        try:
            endpoint = getattr(self.session.rest_api, method.lower())
            self.session.log.debug(f"{self.__class__.__name__}#{method}")
        except AttributeError:
            raise QalxError(f"`{method}` not recognised.")

        success, data = endpoint(*args, data=kwargs, headers=headers)

        if success:
            return data
        self.raise_api_response_error(data)

    def detail_endpoint(self, guid, *args, **kwargs):
        """
        The endpoint for interfacing with a single
        instance of `self.entity_class`
        """
        return f"{self.entity_class.entity_type}/{guid}"

    def list_endpoint(self, *args, **kwargs):
        """
        The endpoint for interfacing with multiple instances of
        `self.entity_class`
        """
        return self.entity_class.entity_type

    @classmethod
    def get_keys_to_save(cls, entity):
        """
        When saving an entity not every key should be saved
        """
        # info & guid are both read only
        # status should only be saved via `update_status` (if available)
        return {k: entity[k] for k in entity if k not in cls._keys_to_exclude}

    @staticmethod
    def _validate(**kwargs):
        """
        Hook for specifying adapter level validation.  Returns kwargs
        so that if we need to mutate any of the kwargs (i.e. ensuring that
        a key not provided by the user has a specific value) then we can
        """
        if kwargs.get("meta", None) is None:
            # meta is optional but must always be sent through as a dict
            kwargs["meta"] = {}
        if not isinstance(kwargs["meta"], dict):
            raise QalxError("`meta` kwarg must be a `dict`")
        return kwargs

    def _process_api_response(
        self, response, entity_class=None, *args, **kwargs
    ):
        """
        Hook for processing the response from the API.  By default
        returns an instance of self.entity_class.  kwargs can be provided
        which will be passed through to entity_class
        :param response: The response dict from the API
        :param entity_class:The entity class that should be instantiated.
        :return: An instance of `entity_class`
        """
        if entity_class is None:
            entity_class = self.entity_class
        return entity_class(
            response, _region=self.session.config.get("REGION"), **kwargs
        )

    def _add_tags(self, entity):
        """
        Helper for applying tags to the entity before adding.  The tags
        need to be added to the individual entity, but can only be applied
        from the session - i.e. the user shouldn't add them manually.
        :param entity: The entity data that is being added
        :return: The entity data.
        """
        if "tags" in entity.keys():
            # Ensures users don't get confused if they try to do
            # `.add(tags={'my': 'tag'}).  We cannot do this in `_validate`
            # because we only want this validation to occur on `add`
            raise QalxError(
                "Tags can only be added to entities via the "
                "session.  Do `session.tags.add(name=<tag_name>, "
                "value=<tag value>)` rather than passing a `tags` "
                "argument to `add`"
            )
        entity["tags"] = self.session.tags
        return entity

    def add(self, **kwargs):
        """
        Adds a new instance of `self.entity_class` or a parent class that is
        provided in `kwargs`.  Provide valid `kwargs` for the object you are
        trying to create

        :return:    An instance of `self.entity_class` or an instance of
                    `parent_class` if this is provided in `kwargs`
        """
        # If parent_class is provided, we return an instance of the parent class
        parent_class = kwargs.pop("parent_class", None)
        if kwargs.pop("_is_validated", False) is False:
            # We may have already validated at this point if we are
            # doing a bulk insert.  This kwarg should be set on the adapter
            # that has done the validation
            kwargs = self._validate(**kwargs)

        self._add_tags(kwargs)

        response = self._process_api_request(
            "post", self.list_endpoint(**kwargs), **kwargs
        )
        if parent_class:
            return self._process_api_response(
                entity_class=parent_class,
                child=self.entity_class,
                response=response,
            )
        return self._process_api_response(response)

    def get(self, guid, *args, **kwargs):
        """
        Gets an instance of `self.entity_class` by the given guid.

        :param guid: The guid of the entity to get
        :return: An instance of `self.entity_class`
        """
        endpoint = self.detail_endpoint(guid=guid, *args, **kwargs)
        resp = self._process_api_request("get", endpoint, *args, **kwargs)
        self.session.log.debug(
            f"get `{self.entity_class.entity_type}` with "
            f"guid `{guid}` with `/{endpoint}`"
        )
        return self._process_api_response(resp)

    def find_one(self, query, **kwargs):
        """
        Method for returning a unique entity. Will return the entity that
        matches the query

        :param query: The mongoDB query
        :type query: dict
        :return: an instance of `self.entity_class`
        :raises QalxMultipleEntityReturned:
        :raises QalxEntityNotFound:
        """
        return self.find(many=False, limit=2, query=query, **kwargs)

    def find(
        self,
        query=None,
        sort=None,
        skip=0,
        limit=25,
        many=True,
        include_session_tags=True,
        **kwargs,
    ):
        """
        Method for listing entities.

        :param query: The optional Mongo query to find entities
        :param sort: The keys to sort by
        :param skip: The number of results to skip (offset) by
        :param limit: How many results should the response be limited to
        :param many: Should many entities be returned or just a single one
        :param include_session_tags:    Should the tags for the session be
                                        included in the query. Default is True
        :param kwargs: kwargs to query by
        :return:
        """
        list_endpoint = self.list_endpoint(**kwargs)

        self.session.log.debug(
            f"find `{self.entity_class.entity_type}` with `/{list_endpoint}`"
        )
        if include_session_tags and self.session.tags:
            # There are tags on this session, modify the query given (if
            # necessary) to include a filter that applies those tags.
            empty_and = {"$and": []}
            if query is None:
                query = empty_and
            elif "$and" not in query:
                query.update(empty_and)
            query["$and"] += self.session.tags._query()

        if query:
            # We need to dump the query to JSON as there could be
            # complex lookups which would get lost when the request is made
            try:
                query = bson_extra.dumps(query)
            except (TypeError, ValueError, OverflowError):
                raise QalxError(
                    "One of the keyword arguments is not JSON serializable"
                )

        resp = self._process_api_request(
            "get",
            list_endpoint,
            sort=sort,
            skip=skip,
            limit=limit,
            query=query,
            **kwargs,
        )
        entities = self._process_api_response(
            resp, entity_class=QalxListEntity, child=self.entity_class
        )

        if many is False:

            # We are expecting only a single entity.
            entities = entities.data
            if len(entities) > 1:
                entities_str = "\n".join([str(q) for q in entities])
                raise QalxMultipleEntityReturned(
                    "Expected one but got "
                    "{}:\n{}".format(len(entities), entities_str)
                )
            elif entities:
                return entities[0]
            else:
                raise QalxEntityNotFound(
                    self.entity_class.entity_type + " not found."
                )
        return entities

    def aggregate(self, aggregate, **kwargs):
        """
        Method for performing an aggregation on data.
        See https://docs.mongodb.com/manual/core/aggregation-pipeline/.  If
        a subset of the data is required then provide the query within the
        aggregation pipeline

        :param aggregate: The aggregation pipeline
        :type aggregate: list
        :return: :class:`~pyqalx.core.entities.entity.AggregationResult`
        """
        try:
            aggregate = bson_extra.dumps(aggregate)
        except (TypeError, ValueError, OverflowError):
            raise QalxError(
                "One of the keyword arguments is not JSON serializable"
            )

        list_endpoint = self.list_endpoint(**kwargs)
        list_endpoint += "/aggregate"
        resp = self._process_api_request(
            "post", list_endpoint, aggregate=aggregate, **kwargs
        )
        return self._process_api_response(
            resp, entity_class=AggregationResult, child=self.entity_class
        )

    def reload(self, entity, *args, **kwargs):
        """
        Reloads the current entity from the API

        :param entity: An instance of `self.entity_class`
        :return: A refreshed instance of `self.entity_class`
        """
        return self.get(entity.guid, *args, **kwargs)

    def save(self, entity, *args, **kwargs):
        """
        Saves `entity` to the database.

        :return: The updated instance of `entity`
        """
        if not entity.get("guid"):
            raise QalxNoGUIDError("No guid.")
        # kwargs is also passed through as it could contain data required
        # for validation by any subclasses (blueprints/files etc)
        entity = self._validate(**entity, **kwargs)
        guid = entity["guid"]

        # If there are tags on the session then use them. Otherwise, fallback
        # to using the tags on the entity. This prevents a session without
        # tags from accidentally removing tags previously set on an entity.
        # If there are no tags on the entity (because they weren't returned via
        # `fields` then we don't update the tags - unless tags have been
        # specified on the session)
        tags_to_save = self.session.tags or entity.get("tags", None)
        if tags_to_save:
            entity["tags"] = tags_to_save

        endpoint = self.detail_endpoint(guid=guid, *args, **kwargs)
        keys_to_save = self.get_keys_to_save(entity=entity)
        self.session.log.debug(
            f"save `{self.entity_class.entity_type}` with "
            f"guid `{guid}` with `/{endpoint}`"
        )

        resp = self._process_api_request("patch", endpoint, **keys_to_save)
        return self._process_api_response(resp)

    def archive(self, entity, *args, **kwargs):
        """
        Archives `entity`

        :param entity: An instance of `self.entity_class`
        :return: The archived instance of `entity`
        """
        guid = entity["guid"]
        detail_endpoint = self.detail_endpoint(guid=guid, *args, **kwargs)
        endpoint = f"{detail_endpoint}/archive"
        self.session.log.debug(
            f"archive `{self.entity_class.entity_type}` with "
            f"guid `{guid}` with `/{endpoint}`"
        )

        resp = self._process_api_request("patch", endpoint)
        return self._process_api_response(resp)

    def delete(self, entity, *args, **kwargs):
        """
        Deletes the given entity from the API

        :param entity: An instance of QalxEntity
        """
        detail_endpoint = self.detail_endpoint(
            guid=entity["guid"], *args, **kwargs
        )
        return self._process_api_request("delete", detail_endpoint)


class QalxUnpackableAdapter(QalxAdapter):
    """
    A qalx adapter for unpackable entities (Set, Group)
    """

    @property
    def kids(self):
        """
        The key that the child entities live in
        """
        return self.child_entity_class.pluralise()

    @staticmethod
    def _child_list_response(child_adapter, query, **kwargs):
        return child_adapter.find(
            query=query, include_session_tags=False, **kwargs
        )

    def _child_list_request(
        self, entity, kids, child_adapter, query_key="guid", child_fields=None
    ):
        """
        Builds the list request for getting any unpackable children. Handles
        chunking of the query and also pagination of the response
        :param entity: The entity that has the children
        :param kids: The details of the kids entities we want to lookup
        :param child_adapter: The child adapter to use
        :param query_key: The query key (defaults to `guid`)
        :return: A list of unpacked child_adapter.entity_class entities
        """
        unpacked_entities = []
        page_size = 100
        _futures = []
        with concurrent.futures.ThreadPoolExecutor() as executor:
            for _page in entity._chunks(kids, chunk_size=page_size):
                # Chunk to avoid hitting maximum request size
                values = list(filter(None, _page))
                query = {query_key: {"$in": values}}
                _futures.append(
                    executor.submit(
                        self._child_list_response,
                        child_adapter=child_adapter,
                        query=query,
                        fields=child_fields,
                        limit=page_size,
                    )
                )
            for future in concurrent.futures.as_completed(_futures):
                try:
                    unpacked_entities += future.result().data
                except QalxAPIResponseError as exc:
                    self.raise_api_response_error(exc)
        return unpacked_entities

    def _kids_for_lookup(self, entity):
        """
        For the given entity, will get the child keys so that an API request
        can be made to obtain them
        :param entity: The entity that has child
        :return: The child keys to send to the API, as a list of values
        """
        return list(entity[self.kids].values())

    def _get_child_entities(self, entity, **kwargs):
        """
        Gets all the child entities for the given entity.  Handles chunking
        the children up into blocks to guard against sending a request that is
        too large.  Also handles the paginated result coming back from the
        `list` endpoint.  Does a `list` lookup for self.child_entity_class
        filtering on `guid` of the children to reduce the amount of queries

        :param entity: The entity to be unpacked
        :return: The child entities as a list
        """
        kids = self._kids_for_lookup(entity)

        child_adapter = getattr(
            self.session, self.child_entity_class.entity_type
        )
        unpacked_entities = self._child_list_request(
            entity=entity, kids=kids, child_adapter=child_adapter, **kwargs
        )
        return unpacked_entities

    def _unpacked_entities_to_valid_children(
        self, entity, unpacked_entities, **kwargs
    ):
        """
        Converts a list of `unpacked_entities` into valid child entities
        for the given entity.  Handles getting any nested children (i.e. items
        on sets when we are unpacking a group).  A two stage process to keep
        queries down in the instance where the same set is on a group multiple
        times

        :param entity: The entity that has kids that should be unpacked
        :param unpacked_entities: The unpacked child entities that need to be
                                  stitched into the entity
        :return: The unpacked entities in the correct format to be assigned to
                 `entity[self.kids]`
        """
        to_return = {}
        fully_unpacked_entities = []
        for child_entity in unpacked_entities:
            # iterate through all the unpacked_entities, getting any further
            # nested children as required
            child_entity_adapter = getattr(
                self.session, child_entity.entity_type
            )
            if hasattr(child_entity_adapter, "child_entity_class"):
                # The child entity has a child adapter of it's own.
                # Attempt unpack. This can occur if we are unpacking
                # a Group as we may need to unpack the Sets on the
                # Group
                child_entity_with_nested = child_entity_adapter._attempt_unpack(
                    child_entity, **kwargs
                )
                # At this point this given `child_entity` will have all
                # the nested children within fully unpacked.
                fully_unpacked_entities.append(child_entity_with_nested)
            else:
                # This is the bottom of the unpacking hierarchy.
                fully_unpacked_entities.append(child_entity)

        for key, child_guid in entity[self.kids].items():
            # Iterate through the key/values on the children of this entity
            # and update the appropriate key from the fully_unpacked_entities.
            for child_entity in fully_unpacked_entities:
                if child_guid == child_entity.guid:
                    to_return[key] = child_entity
        return to_return

    def _attempt_unpack(self, entity, **kwargs):
        """
        For the given entity will attempt
        to unpack the child_entity data that it contains

        :entity: The entity to be unpack
        :return: The entity instance with the child_entity objects unpacked
        """

        if self.kids in entity:
            # `self.kid` might not be there if we have specified a subset
            # of fields
            unpacked_entities = self._get_child_entities(entity, **kwargs)
            entity[self.kids] = self._unpacked_entities_to_valid_children(
                entity=entity, unpacked_entities=unpacked_entities, **kwargs
            )
        return entity

    def _pack_kids_for_request(self, kwargs):
        """
        Packs the kids into a format that they need to be in
        for sending to the API
        :param kwargs: The kwargs sent to the `add` method
        :return: kwargs with packed kids
        """
        kids = {key: str(kid.guid) for key, kid in kwargs[self.kids].items()}
        kwargs[self.kids] = kids
        return kwargs

    def get_keys_to_save(self, entity):
        """
        Unpackable entities will have unpacked data stored on them that have
        to be submitted as packed data
        """
        keys_to_save = super(QalxUnpackableAdapter, self).get_keys_to_save(
            entity=entity
        )
        if self.kids in keys_to_save.keys():
            keys_to_save[self.kids] = self.get_kids_keys_to_save(
                keys_to_save[self.kids]
            )
        return keys_to_save

    @staticmethod
    def get_kids_keys_to_save(kids_to_save):
        """
        Gets the keys of the kids that we want to save.  Ensures that if the
        keys are in an packed state we don't attempt to unpack them.
        :param kids_to_save: A dict of the kids we want to save
        :type kids_to_save: dict
        :return: A dict of {key:guid} of the kids we want to save
        """
        return {
            k: str(i["guid"]) if isinstance(i, dict) else i
            for k, i in kids_to_save.items()
        }

    @staticmethod
    def cache_original_kids(kids):
        """
        When saving the API only wants to know about the guids of the kids so
        we need to cache the originals
        until after the save
        :param kids: The kids on this entity
        :type kids: dict
        :return: A dictionary of the kids
        """
        return {k: v for k, v in kids.items()}

    def get(self, guid, child_fields=None, unpack=True, *args, **kwargs):
        """
        Gets the entity for the given `guid` and unpacks it if specified.
        Provide `child_fields` to restrict the fields returned on unpacked
        children

        :param guid: The `guid` of the entity to get
        :type guid: str
        :param child_fields: A list of fields that should be returned from
                             child entities
        :type child_fields: list
        :param unpack: Should the child entities be unpacked. Defaults to True
        :type unpack: bool
        :return: An unpacked entity
        """
        entity = super(QalxUnpackableAdapter, self).get(
            guid=guid, *args, **kwargs
        )
        if unpack is True:
            entity = self._attempt_unpack(entity, child_fields=child_fields)
        return entity

    def reload(self, entity, unpack=True, *args, **kwargs):
        entity = super(QalxUnpackableAdapter, self).reload(
            entity, unpack=unpack, *args, **kwargs
        )
        return entity

    def find_one(self, query, unpack=True, **kwargs):
        entity = super(QalxUnpackableAdapter, self).find_one(
            query, unpack=unpack, **kwargs
        )
        return entity

    def find(self, many=True, child_fields=None, *args, **kwargs):
        """
        Return multiple packed entities from the API

        :param many: Are many results expected or just one?
        :type many: bool
        :param child_fields: A list of fields that should be returned from
                             child entities
        :type child_fields: list
        :return: A list of entities
        """
        entities = super(QalxUnpackableAdapter, self).find(
            many=many, *args, **kwargs
        )
        unpack = kwargs.get("unpack", True)
        if unpack is True and many is False:
            # Only unpack a list view if we have a single entity
            entities = self._attempt_unpack(entities, child_fields=child_fields)
        return entities

    def save(self, entity, *args, **kwargs):
        """
        When saving an unpacked entity the kids need to be packed. To save
        having to unpack them again just save the original kids and replace
        them with the packed values after saving

        :param entity: An instance of `self.entity_class`
        :return: An updated instance of `self.entity_class`
        """
        has_kids_key = self.kids in entity.keys()
        if has_kids_key:
            original_kids = self.cache_original_kids(entity[self.kids])

        entity = super(QalxUnpackableAdapter, self).save(
            entity=entity, *args, **kwargs
        )
        if has_kids_key:
            entity[self.kids] = original_kids
        else:
            entity = self._attempt_unpack(entity=entity)
        return entity

    def cache_kids(self, kwargs):
        """
        Gets the kids from the given kwargs

        :param kwargs:a dict with a `self.kids` key
        :type kwargs: dict
        :return:The value of the `self.kids` key or None
        """
        return kwargs.get(self.kids, None)

    def save_kids_on_entity(self, entity, kids):
        """
        Updates the given `self.kids` key on the entity with the given kids

        :param entity: An entity to be updated
        :param kids: The kids to update the entity with
        """
        entity[self.kids] = kids

    def add(self, **kwargs):
        # Need to validate before we pack the kids up
        kwargs = self._validate(**kwargs)
        # Cache the kids to reassign them to the newly created entity

        cached_kids = self.cache_kids(kwargs)
        kwargs = self._pack_kids_for_request(kwargs)
        entity = super(QalxUnpackableAdapter, self).add(
            _is_validated=True, **kwargs
        )
        if cached_kids is not None:
            self.save_kids_on_entity(entity, cached_kids)
        return entity


class QalxSignalAdapter(QalxAdapter):
    """
    Bots and Workers have signals that are used to determine when to stop
    processing data.  This class provides that shared functionality
    """

    def get_signal(self, entity, *args, **kwargs):
        """
        Gets just the `signal` field from the entity and then parses that
        into the `signal_class`
        """
        entity = self.get(guid=entity["guid"], fields="signal", *args, **kwargs)
        signal = self.signal_class(entity)
        return signal

    def terminate(self, entity, *args, **kwargs):
        """
        Updates the entity with a terminate signal
        """
        guid = entity["guid"]
        endpoint = self.detail_endpoint(guid=guid, **kwargs)
        self.session.log.debug(
            f"terminate `{self.entity_class.entity_type}` with guid `{guid}` with `/{endpoint}`"  # noqa
        )
        signal = self.signal_class._terminate_signal(*args, **kwargs)

        self._process_api_request("patch", endpoint, signal=signal)

    def _stop_or_resume(self, entity, stop, **kwargs):
        guid = entity["guid"]
        endpoint = self.detail_endpoint(guid=guid, **kwargs)
        signal = self.signal_class._stop_signal(stop)
        self.session.log.debug(
            f"signal `{signal}` on `{self.entity_class.entity_type}` with guid `{guid}` with `/{endpoint}`"  # noqa
        )
        self._process_api_request("patch", endpoint, signal=signal)

    def stop(self, entity, **kwargs):
        """
        Updates the entity with a stop signal
        """
        self._stop_or_resume(entity, stop=True, **kwargs)

    def resume(self, entity, **kwargs):
        """
        Updates the entity with a resume signal
        """
        self._stop_or_resume(entity, stop=False, **kwargs)


class QalxNamedEntityAdapter(QalxAdapter):
    """
    Certain QalxEntities have a `name` field that can be used to look up the
    entity.
    """

    def get_by_name(self, name, **kwargs):
        """gets a single entity by name

        :param name: name of entity
        :type name: str
        :param kwargs: Keyword arguments to be used in the find method
        :return: instance of `self.entity_class`
        :raises QalxMultipleEntityReturned:
        :raises QalxEntityNotFound:
        """
        return self.find_one(query={"name": name}, **kwargs)

    def get_or_create(self, name, meta=None, **kwargs):
        """
        Gets an entity by the given name or creates it if it doesn't exist

        :param name: The `name` that to use to get or create the entity by
        :type name: str
        :param meta: metadata about the entity
        :param kwargs: Other kwargs on the entity to use when creating
        :return: instance of `self.entity_class`
        """
        try:
            return self.get_by_name(name=name)
        except QalxEntityNotFound:
            return self.add(name=name, meta=meta, **kwargs)


class QalxValidateBlueprintAdapter(QalxAdapter):
    """
    Certain entity types can be validated against a preexisting Blueprint
    when they are created.
    """

    def _validate(self, blueprint_name=None, **kwargs):
        """
        When adding entities that inherit from this class, pyqalx will get
        the blueprint based on the `blueprint_name` and then ensure that
        the schema validates correctly against the data that you are creating
        the entity with

        :param blueprint_name: the name of the blueprint
        :type blueprint_name: str, None
        :param get_blueprint: Should this method retrieve the blueprint?
        :type get_blueprint: bool, False
        :return: A valid `self.entity_class` instance
        :raises: jsonschema.ValidationError
        """
        validated_data = super(QalxValidateBlueprintAdapter, self)._validate(
            **kwargs
        )
        if blueprint_name is not None:
            blueprint = validated_data.pop("blueprint", None)
            if blueprint is None:
                blueprint = self._get_blueprint(name=blueprint_name)
            blueprint.validate(validated_data)
        return validated_data

    def _get_blueprint(self, name):
        """
        Helper method for getting the blueprint instance
        :param name: The name of the blueprint
        :type name: str
        :return: A `~entities.blueprint.Blueprint` instance
        """
        blueprint = self.session.blueprint.get_by_name(name=name)
        return blueprint

    def add(self, blueprint_name, **kwargs):
        """
        When adding entities that inherit from this class, pyqalx will get
        the blueprint based on the `blueprint_name` and then ensure that
        the schema validates correctly against the data that you are creating
        the entity with

        :param blueprint_name: the name of the blueprint
        :type blueprint_name: str, None
        :return: A valid `self.entity_class` instance
        :raises: jsonschema.ValidationError
        """
        return super(QalxValidateBlueprintAdapter, self).add(
            blueprint_name=blueprint_name, **kwargs
        )

    def save(self, entity, blueprint_name=None, *args, **kwargs):
        """
        When saving entities that inherit from this class, pyqalx will get
        the blueprint based on the `blueprint_name` and then ensure that
        the schema validates correctly against the data that you are updating
        the entity with

        :param entity: The entity instance that you are saving
        :param blueprint_name: the name of the blueprint
        :type blueprint_name: str, None
        :return: A valid `self.entity_class` instance
        :raises: jsonschema.ValidationError
        """
        return super(QalxValidateBlueprintAdapter, self).save(
            entity, blueprint_name=blueprint_name, **kwargs
        )


class QalxFileAdapter(QalxAdapter):
    """
    Adapter that allows files to be uploaded to an entity
    """

    def __init__(self, *args, **kwargs):
        self._cached_data = None
        super().__init__(*args, **kwargs)

    def _process_api_response(self, *args, **kwargs):
        """
        Ensures that if there is a KEYFILE configured that this is present
        on all instances of `self.entity_class` to provide seamless decryption
        of files when downloading
        :param response: The response dict from the API
        :return: An instance of `self.entity_class`
        """
        kwargs["keyfile"] = self.session.config.get("KEYFILE")

        # Handle `data` dict manipulation if stored as file
        try:
            if args[0]["file"]["as_file"]:
                if self._cached_data:
                    # POST or PATCH/PUT
                    args[0]["data"] = self._cached_data
                else:
                    args[0]["data"] = self._retrieve_data_file(args[0]["file"])
        except KeyError:
            try:
                if args[0]["query"]:
                    # Handle response of LIST (find)
                    with ThreadPoolExecutor() as pool:
                        futures = {
                            pool.submit(
                                self._retrieve_data_file, item["file"]
                            ): item["guid"]
                            for item in args[0]["data"]
                        }
                        for i, obj in enumerate(as_completed(futures)):
                            if obj.result():
                                # Will only assign to `data` if the item is a
                                # data file item
                                args[0]["data"][i]["data"] = obj.result()
            except (IndexError, KeyError):
                pass
        except IndexError:
            pass
        return super(QalxFileAdapter, self)._process_api_response(
            *args, **kwargs
        )

    @staticmethod
    def _retrieve_data_file(file_dict):
        if file_dict["as_file"]:
            # Only return content if the item is a data is stored as file
            url = file_dict["url"]
            file_retrieved = False
            sleep_time = 3
            while not file_retrieved:
                resp = requests.get(url)
                if resp.status_code == HTTPStatus.NOT_FOUND:
                    if sleep_time > 0:
                        time.sleep(sleep_time)
                        sleep_time -= 1
                    else:
                        break
                else:
                    file_retrieved = True
                    break
            if file_retrieved:
                return bson_extra.loads(resp.content.decode())
            else:
                # TODO: Improve error message
                raise QalxFileError("Could not retrieve file data from S3.")

    def _validate(self, source, file_name, as_file=False, **kwargs):
        """
        Ensures that the `source` and/or `file_name` and `as_file are present
        and correct
        :param source: The path to the file or the file stream
        :param file_name: The name of the file
        :param as_file: Whether the `data` dict should be stored in S3 rather
        than in Mongo
        :return: kwargs
        """
        if source is not None:
            # Only validate the file if it's specified in the
            # implementing adapter
            file = {}
            if as_file:
                # Handle `data` saved as file
                if kwargs.get(self.entity_class._file_key):
                    # Saving an existing item
                    file = kwargs.get(self.entity_class._file_key)
                else:
                    # Adding a new item
                    file = {
                        "name": f"item_data_{uuid.uuid4()}.bson",
                        "as_file": True,
                    }
            else:
                is_filestream = self.session.rest_api.is_filestream(source)
                if is_filestream:
                    if not file_name:
                        raise QalxFileError(
                            "You must supply a file name when supplying"
                            " a file stream"
                        )
                    file["name"] = file_name
                else:
                    if os.path.isfile(source):
                        file_path = os.path.abspath(source)
                        filename = (
                            file_name
                            if file_name
                            else os.path.basename(file_path)
                        )
                        file["name"] = filename
                    else:
                        raise QalxFileError(f"Invalid file `{source}`")
            key_file = self.session.config.get("KEYFILE")
            if key_file is not None:
                # The user can suppress encryption on a per-case basis, by
                # passing `encrypt=False`
                if kwargs["encrypt"]:
                    file["keyfile"] = os.path.basename(key_file)
            kwargs[self.entity_class._file_key] = file
        return super(QalxFileAdapter, self)._validate(**kwargs)

    def add(
        self,
        source,
        file_name,
        data=None,
        as_file=False,
        upload=True,
        encrypt=True,
        **kwargs,
    ):
        """
        When a user creates an entity with a file pyqalx will upload the file
        automatically for you after the entity itself has been created.  Pass
        `upload=False` through to stop this from happening and upload the file
        yourself

        :param source: The input file path or stream
        :param file_name: The optional file name
        :param data: Data to store on this item
        :param as_file: Whether the `data` dict should be stored in S3 rather
        than in Mongo
        :param upload: Whether the file should be automatically uploaded or not
        :type upload: bool
        :param encrypt: Whether the file should be automatically encrypted if
         the KEYFILE is present
        :type encrypt: bool
        :param as_file: Whether the `data` dict should be stored in S3 rather
        than in Mongo
        :return: A valid entity
        """
        if data is None:
            data = {}
        source = self._handle_data_as_file(data, source, as_file)
        entity = super(QalxFileAdapter, self).add(
            data=data,
            source=source,
            file_name=file_name,
            encrypt=encrypt,
            as_file=as_file,
            **kwargs,
        )
        if upload and source is not None:
            entity = self._upload_file(entity=entity, source=source)
        return entity

    def save(
        self,
        entity,
        source,
        file_name,
        upload=True,
        encrypt=True,
        as_file=False,
        *args,
        **kwargs,
    ):  # noqa
        """
        Saves an entity - optionally updating any file data.

        :param entity: The entity that is being saved
        :param source: The input file path or stream
        :param file_name: The optional file name
        :type file_name: str
        :param upload: Whether the file should be automatically uploaded or not
        :type upload: bool
        :param encrypt: Whether the file should be automatically encrypted if
         the KEYFILE is present
        :type encrypt: bool
        :param as_file: Whether the `data` dict should be stored in S3 rather
        than in Mongo
        :return: A valid entity
        """
        # If item data is already saved as file, then keep it that way
        as_file = as_file or (
            entity.file.get("as_file") if entity.get("file") else False
        )
        source = self._handle_data_as_file(
            entity.get("data", {}), source, as_file
        )
        entity = super(QalxFileAdapter, self).save(
            entity=entity,
            source=source,
            file_name=file_name,
            encrypt=encrypt,
            as_file=as_file,
            **kwargs,
        )
        if source is not None:
            if upload:
                entity = self._upload_file(entity=entity, source=source)  # noqa
            else:
                # Entity is returned with the put_url so that the user can
                # upload the file manually
                return entity
        # Ensure "put_url" is removed even if we are not changing the file
        entity[self.entity_class._file_key].pop("put_url", None)
        return entity

    def _upload_file(self, entity, source):
        # Only upload if everything was OK with the item creation
        # and the user wants to upload the file automatically.

        s3_resp_ok, s3_data = self.session.rest_api._upload_to_s3(
            entity, source, self.entity_class._file_key
        )
        if s3_resp_ok is False:
            # The S3 upload failed, return the failed S3 response in place
            # of the response from the api.  Any implementing client will
            # have to handle tidying up any incorrect data.
            self.raise_api_response_error(s3_data)
        else:
            # Always pop the `put_url` off if it's a success
            entity[self.entity_class._file_key].pop("put_url")
        return entity

    def _handle_data_as_file(self, data, source, as_file):
        if as_file:
            # Need to raise error here if required
            if source is not None:
                raise QalxFileError(
                    "Cannot specify a file `source` and have `as_file` = True"
                    "together"
                )
            source = io.BytesIO()
            source.write(bson_extra.dumps(data).encode())
            source.seek(0)

            # 'Cache' the data dict in memory so that it can be returned to the
            # user without saving it to Mongo or having to retrieve it from S3
            self._cached_data = copy.deepcopy(data)
            data.clear()
        return source


class QalxUpdateStatusAdapter(QalxAdapter):
    def update_status(self, entity, status, **kwargs):
        """
        Updates the status of an entity

        :param entity: The entity that is being updated
        :param status: The status to update to
        :type status: str
        """
        guid = entity["guid"]
        endpoint = self.detail_endpoint(guid=guid, **kwargs)
        self.session.log.debug(
            f"update status `{self.entity_class.entity_type}` with guid `{guid}` with `/{endpoint}`"  # noqa
        )
        status_dict = {"message": status}
        self._process_api_request("patch", endpoint, status=status_dict)


class QalxUpdateStateAdapter(QalxAdapter):
    def update_state(self, entity, state: ValidStates, **kwargs):
        """
        Helper to update the state of a bot or worker. The API handles
        validation of the `state` arg.

        :param entity: The entity that is being updated
        :param state: The state to update to
        """
        endpoint = self.detail_endpoint(guid=entity["guid"], **kwargs)
        self._process_api_request("patch", endpoint, state=state.value)



from .core import BasePyQalxAPI, PyQalxAPIException


class PyQalxAPI(BasePyQalxAPI):
    """
    The main interface between adapters and the REST API.

    Usage:

    >>> from pyqalx import QalxSession
    >>> qalx = QalxSession()
    >>> qalx.rest_api
    """

    def get(self, endpoint, **kwargs):
        """
        Performs a GET request to the specified endpoint

        :param endpoint: The endpoint you want to query
        :type endpoint: str
        :param kwargs: Any arguments you want to pass to the request.
        :type kwargs: dict
        :return: tuple of (bool, dict).  The first element is whether the response
                 was successful, the second element is the response data
        """
        url = self._get_url(endpoint)
        if kwargs.get("data", None) and kwargs.get("params", None):
            # They have supplied params and data.  Raise an exception as we
            # can't safely merge them
            raise PyQalxAPIException(
                "qalx GET parameters must be supplied"
                "using the `params` parameter for API"
                "consistency"
            )
        data_removed = kwargs.pop("data", None)
        if data_removed and kwargs.get("params") is None:
            # For API consistency the `params` for a query are stored on `data`
            # but need to be moved to `params` for the request. Any user using
            # the transport layer on its own can send either `data` or `params`.
            # The querystring cannot be in the request body
            kwargs["params"] = data_removed
        return self._build_request(url=url, method="GET", **kwargs)

    def post(self, endpoint, **kwargs):
        """
        Performs a POST request to the specified endpoint

        :param endpoint: The endpoint you want to query
        :type endpoint: str
        :param kwargs: Any arguments you want to pass to the request.
        :type kwargs: dict
        :return: tuple of (bool, dict).  The first element is whether the response
                 was successful, the second element is the response data
        """
        return self._build_non_safe_request(
            method="POST", endpoint=endpoint, **kwargs
        )

    def patch(self, endpoint, **kwargs):
        """
        Performs a PATCH request to the specified endpoint

        :param endpoint: The endpoint you want to query
        :type endpoint: str
        :param kwargs: Any arguments you want to pass to the request.
        :type kwargs: dict
        :return: tuple of (bool, dict).  The first element is whether the response
                 was successful, the second element is the response data
        """
        return self._build_non_safe_request(
            method="PATCH", endpoint=endpoint, **kwargs
        )

    def put(self, endpoint, **kwargs):
        """
        Performs a PUT request to the specified endpoint

        :param endpoint: The endpoint you want to query
        :type endpoint: str
        :param kwargs: Any arguments you want to pass to the request.
        :type kwargs: dict
        :return: tuple of (bool, dict).  The first element is whether the response
                 was successful, the second element is the response data
        """
        return self._build_non_safe_request(
            method="PUT", endpoint=endpoint, **kwargs
        )

    def delete(self, endpoint, **kwargs):
        """
        Performs a DELETE request to the specified endpoint

        :param endpoint: The endpoint you want to query
        :type endpoint: str
        :param kwargs: Any arguments you want to pass to the request.
        :type kwargs: dict
        :return: tuple of (bool, dict).  The first element is whether the response
                 was successful, the second element is the response data
        """
        url = self._get_url(endpoint)
        return self._build_request(url=url, method="DELETE")



class FactoryPlanDebug(dict):
    """
    A subclass of dict that allows us to specify the default values to use
    for the various debug settings available to a factory
    """

    DEMOLISH_ON_FAILURE_KEY = "demolish_on_failure"

    def __init__(self, *args, **kwargs):
        super(FactoryPlanDebug, self).__init__(*args, **kwargs)

        for (
            _debug_key,
            _debug_configuration,
        ) in self._debug_configuration().items():
            self.setdefault(_debug_key, _debug_configuration["default"])
            # Set it as an attribute for easier lookups
            setattr(self, f"_{_debug_key}", self[_debug_key])

    @classmethod
    def _debug_configuration(cls):
        """
        The main source of truth for factory debugging.  Keys in here can be
        specified on a factory plan.  Keys not provided on the factory plan
        will default to the value here.
        Usage:
        # factory-plan.yaml
        factory:
           debug:
              demolish_on_failure: false
        """
        return {
            # Should we demolish the factory on build failure?  Ensure we
            # do this by default
            cls.DEMOLISH_ON_FAILURE_KEY: {"default": True}
        }

    @property
    def _demolish_on_failure_msg(self):
        """
        Helper to show a consistent message in the various places that we
        inform the user that demolishing isn't happening
        """
        return (
            "Resources have not been deleted because "
            f"`{self.DEMOLISH_ON_FAILURE_KEY}` is set to false. These will "
            "have to be tidied up manually using the "
            "`qalx factory-demolish` command. "
            "See the logs for more details"
        )



import os
import sys

import click
from click import ClickException, Choice

from pyqalx import QalxSession
from pyqalx.config import UserConfig, BotConfig
from pyqalx.core.errors import (
    QalxConfigProfileNotFound,
    QalxConfigFileNotFound,
    QalxConfigError,
    QalxAPIResponseError,
    QalxError,
)
from qalxcli import cli_types
from qalxcli.utils import generate_key_file


@click.group(chain=True)
@click.option(
    "-u",
    "--user-profile",
    help="User profile name in .qalx.",
    default=UserConfig.default,
)
@click.option(
    "-b",
    "--bot-profile",
    help="Bot profile name in .bots.",
    default=BotConfig.default,
)
@click.version_option()
@click.pass_context
def qalx(ctx, user_profile, bot_profile):
    """Command line interface to qalx."""
    ctx.ensure_object(dict)
    ctx.obj["USER_PROFILE"] = user_profile
    ctx.obj["BOT_PROFILE"] = bot_profile
    sys.path.append(os.getcwd())


@qalx.command("configure")
@click.option("--user/--no-user", default=True)
@click.option("--bot/--no-bot", default=True)
@click.argument("extra", nargs=-1, type=cli_types.QALX_DICT)
@click.pass_context
def configure(ctx, user, bot, extra):
    """
    Configures the .qalx and .bots config files.
    Usage:

    `qalx configure` - configures user and bot default profiles

    `qalx --user-profile=dev configure` - configures dev user profile and
    default bot profile

    `qalx --bot-profile=dev configure --no-user` - only configures bot dev
    profile.  Doesn't configure user profile

    `qalx configure --no-bot customkey=customvalue customkey2=customvalue2`
    - configures user profile with two extra key/values pairs on the config
    """

    if not user and not bot:
        raise ClickException(
            "`user` or `bot` must be specified otherwise no "
            "config files can be created.  Either call "
            "without any arguments or only use a "
            "single `--no-user` or `--no-bot` switch"
        )

    user_profile = ctx.obj["USER_PROFILE"]
    bot_profile = ctx.obj["BOT_PROFILE"]
    check_existing = []

    if user:
        check_existing.append((UserConfig, user_profile, "user"))
    if bot:
        check_existing.append((BotConfig, bot_profile, "bot"))

    for ConfigClass, profile_name, config_type in check_existing:
        # If the profiles already exist we just exit.  It's up to the user
        # to make any changes manually to the profile file - otherwise we risk
        # parsing the file incorrectly and overwriting data the user wants to
        # keep
        try:
            ConfigClass().from_inifile(profile_name)
            config_path = ConfigClass.config_path()
            raise ClickException(
                f"`{profile_name}` profile for config stored "
                f"at `{config_path}` "
                f"already exists.  To make changes you must "
                f"edit the config file directly or specify a "
                f"`--no-{config_type}` flag"
            )
        except (QalxConfigProfileNotFound, QalxConfigFileNotFound):
            # QalxConfigProfileNotFound: The specific profile could not be
            #                            found in the given ConfigClass file
            # QalxConfigFileNotFound: The specific file could not be found
            #                         for the given ConfigClass
            pass

    config_items = {k.upper(): v for d in extra for k, v in d.items()}

    if user:
        # Get the token
        value = click.prompt("qalx Token", type=str)
        config_items["TOKEN"] = value

        region = click.prompt(
            "region",
            default=UserConfig.DEFAULT_REGION,
            type=Choice(choices=UserConfig.REGION_MAPPING["aws"].keys()),
        )
        config_items["REGION"] = UserConfig.REGION_MAPPING["aws"][region]

        # Get the key file for encryption
        key_file = config_items.get("KEYFILE")
        if key_file is None:
            if click.confirm("Do you have an encryption key file?"):
                key_file = click.prompt(
                    "Enter full path to encryption key file"
                )
            elif click.confirm(
                "Do you want to generate an encryption key file?"
            ):
                key_file = generate_key_file()
            if key_file is not None:
                config_items["KEYFILE"] = key_file
        try:
            UserConfig.configure(user_profile, config_items)
        except QalxConfigError as exc:
            raise ClickException(exc)
    if bot:
        try:
            BotConfig.configure(bot_profile)
        except QalxConfigError as exc:
            raise ClickException(exc)


@qalx.command("generate-encryption")
def generate_encryption():
    """
    Generates an encryption key file for the user prints out the path if it does
    so successfully
    """
    key_file_path = generate_key_file()
    click.echo(
        f"\nThe encryption key file has been generated in:\n{key_file_path}\n"
        + "The qalx configuration file must be updated manually for the new "
        + "file to be used, by adding the following line under the appropriate "
        + f"profile:\n\nKEYFILE={key_file_path}"
    )


@qalx.command("download")
@click.argument("guid")
@click.option(
    "-d",
    "--directory",
    type=cli_types.DIR_PATH_EXISTING,
    help="The directory where the file will be downloaded.",
    default=None,
)
@click.option(
    "-e",
    "--entity-type",
    type=cli_types.FILE_ENTITY_TYPE,
    help="The type of entity to download a file for.",
    default="item",
)
@click.pass_context
def download(ctx, guid, directory, entity_type):
    """
    Downloads a file for the specified entity type and guid. The download
    directory can also be specified, otherwise the default location from the
    qalx config file will be used
    """
    qalx_session = QalxSession(profile_name=ctx.obj["USER_PROFILE"])
    if directory is None:
        directory = qalx_session.config["DOWNLOAD_DIR"]
        # The directory from the config might not exist
        if not os.path.exists:
            os.mkdir(directory)
    try:
        entity = getattr(qalx_session, entity_type).get(guid)
    except QalxAPIResponseError as exc:
        msg = f"{entity_type} with guid: {guid} not found:\n\n{exc}"
        raise ClickException(msg)
    try:
        entity.save_file_to_disk(directory)
    except QalxError as exc:
        raise ClickException(exc)



"""
Functionality related to encryption of files in qalx
"""

import os

import appdirs
from cryptography.fernet import Fernet

from ..core.utils import APPNAME, APPAUTHOR


class QalxEncrypt:
    """
    Provides functionality to encrypt and decrypt files
    """

    DEFAULT_KEY_PATH = os.path.join(
        appdirs.user_data_dir(APPNAME, APPAUTHOR), "qalx.key"
    )
    encryption_class = Fernet

    def __init__(self, key_file):
        self.key_file = key_file

    def load(self):
        """Read key_file in"""
        with open(self.key_file, "rb") as key_file:
            return self.encryption_class(key_file.read())

    def encrypt(self, file_bytes):
        """
        Encrypt file data.
        :param file_bytes:  The bytes to encrypt
        :return The encrypted bytes
        """
        return self.load().encrypt(file_bytes)

    def decrypt(self, file_bytes):
        """
        Decrypt file data.
        :param file_bytes:  The bytes to decrypt
        :return The decrypted bytes
        """
        return self.load().decrypt(file_bytes)

    @staticmethod
    def generate_key():
        """Return an encryption key"""
        return Fernet.generate_key()

    @classmethod
    def save_key(cls, key_path=None):
        """
        Saves a new encryption key to the provided file location

        :param key_path:    The path to the file where the new key will be saved
                            If None is provided, it defaults to DEFAULT_KEY_PATH
        """
        if key_path is None:
            key_path = cls.DEFAULT_KEY_PATH
        key = cls.generate_key()
        with open(key_path, "wb") as key_file:
            key_file.write(key)
